from celery.app import default_app

from django.core.cache import cache
from django.test import TransactionTestCase, override_settings

from taskmonitor.models import TaskLog

TASK_LOGS_PATH = "taskmonitor.core.task_logs"


@default_app.task
def normal_task():
    pass


@default_app.task
def special_task():
    """Task returns an object, which is not JSON serializable."""
    return Dummy()


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestTaskLogCreation(TransactionTestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_should_create_entry_for_normal_task(self):
        # when
        normal_task.delay()
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_name="taskmonitor.tests.integration.test_tasklogs.normal_task",
                state=TaskLog.State.SUCCESS,
            ).exists()
        )

    def test_should_create_entry_for_task_which_result_can_not_be_serialized(self):
        # when
        special_task.delay()
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_name="taskmonitor.tests.integration.test_tasklogs.special_task",
                state=TaskLog.State.SUCCESS,
            ).exists()
        )


class Dummy:
    pass



"""Views for Task Monitor."""

import csv

from django.conf import settings
from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.http import Http404, JsonResponse, StreamingHttpResponse
from django.shortcuts import redirect, render

from allianceauth import NAME as site_header
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from taskmonitor import __title__, tasks
from taskmonitor.app_settings import (
    TASKMONITOR_DATA_MAX_AGE,
    TASKMONITOR_REPORTS_MAX_TOP,
)
from taskmonitor.core import cached_reports, celery_queues
from taskmonitor.helpers import Echo
from taskmonitor.models import TaskLog, TaskStatistic

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@staff_member_required
def admin_taskmonitor_download_csv(request) -> StreamingHttpResponse:
    """Return all tasklogs as CSV file for download."""
    queryset = TaskLog.objects.order_by("pk")
    model = queryset.model
    exclude_fields = ("traceback", "args", "kwargs", "result", "current_queue_length")

    logger.info("Preparing to export the task log with %s entries.", queryset.count())

    fields = [
        field
        for field in model._meta.fields + model._meta.many_to_many
        if field.name not in exclude_fields
    ]
    pseudo_buffer = Echo()
    writer = csv.writer(pseudo_buffer, delimiter=";")
    return StreamingHttpResponse(
        (writer.writerow(row) for row in queryset.csv_line_generator(fields)),
        content_type="text/csv",
        headers={"Content-Disposition": 'attachment; filename="tasklogs.csv"'},
    )


@login_required
@staff_member_required
def admin_taskmonitor_reports(request):
    """Show the reports page."""
    report = cached_reports.report("tasks_basics")
    report_data = report.data()
    total_runs = report_data["total_runs"]
    oldest_date = report_data["oldest_date"]
    newest_date = report_data["newest_date"]
    context = {
        "title": "Reports",
        "site_header": site_header,
        "cl": {"opts": TaskLog._meta},
        "data_max_age": TASKMONITOR_DATA_MAX_AGE,
        "debug_mode": settings.DEBUG,
        "total_runs": total_runs,
        "oldest_date": oldest_date,
        "newest_date": newest_date,
        "MAX_TOP": TASKMONITOR_REPORTS_MAX_TOP,
        "last_update_at": report.last_update_at(),
    }
    return render(request, "admin/taskmonitor/report/index.html", context)


@login_required
@staff_member_required
def admin_taskmonitor_reports_clear_cache(request):
    """Reload the reports page with cleared cache."""
    cached_reports.clear_cache()
    return redirect("taskmonitor:admin_taskmonitor_reports")


@login_required
@staff_member_required
def admin_taskmonitor_reports_recalculation(request):
    """Start the reports recalculation."""
    tasks.refresh_cached_data.apply_async(priority=tasks.DEFAULT_TASK_PRIORITY)
    messages.info(
        request,
        (
            "Reports are being recalculated, which can take a while. "
            "Please reload this page in a minute to see the update."
        ),
    )
    return redirect("taskmonitor:admin_taskmonitor_reports")


@login_required
@staff_member_required
def admin_taskmonitor_report_json(request, report_name: str):
    """Render report in JSON."""
    use_cache = request.GET.get("use_cache") != "false"
    try:
        data = {"data": cached_reports.report_data(report_name, use_cache=use_cache)}
    except KeyError:
        raise Http404(f'No report with name: "{report_name}"') from None
    return JsonResponse(data)


@login_required
@staff_member_required
def admin_taskmonitor_report_html(request, report_name: str):
    """Render report in HTML."""
    use_cache = request.GET.get("use_cache") != "false"
    try:
        data = cached_reports.report_data(report_name, use_cache=use_cache)
    except KeyError:
        raise Http404(f'No report with name: "{report_name}"') from None
    disable_percent = request.GET.get("disable_percent") == "yes"
    context = {"data": data, "disable_percent": disable_percent}
    return render(
        request, "admin/taskmonitor/report/render_table_partial.html", context
    )


@login_required
@staff_member_required
def admin_taskmonitor_report_debug(request, report_name: str):
    """Render report in HTML."""
    try:
        cached_reports.report_data(report_name, use_cache=False)
    except KeyError:
        raise Http404(f'No report with name: "{report_name}"') from None
    context = {
        "title": "DEBUG Reports",
        "cl": {"opts": TaskLog._meta},
        "report_name": report_name,
    }
    return render(request, "admin/taskmonitor/report/debug.html", context)


@login_required
@staff_member_required
def admin_queued_task_purge(request):
    """Purge the task queue."""
    queue_length = celery_queues.queue_length()
    celery_queues.clear_tasks()
    messages.info(request, f"Purged queue with {queue_length:,} tasks.")
    return redirect("admin:taskmonitor_queuedtask_changelist")


@login_required
@staff_member_required
def admin_queued_task_clear_cache(request):
    """Clear the cache for queued tasks."""
    celery_queues.tasks_cache.clear()
    return redirect("admin:taskmonitor_queuedtask_changelist")


@login_required
@staff_member_required
def admin_taskmonitor_statistics_clear_cache(request):
    """Reload the reports page with cleared cache."""
    TaskStatistic.objects.clear_cache()
    return redirect("admin:taskmonitor_taskstatistic_changelist")



from io import StringIO
from unittest.mock import patch

from django.core.management import call_command
from django.test import TestCase

from taskmonitor.core.celery_queues import QueuedTaskShort
from taskmonitor.management.commands import taskmonitorctl
from taskmonitor.models import TaskLog

from .factories import QueuedTaskRawFactory, TaskLogFactory

MODULE_PATH = "taskmonitor.management.commands.taskmonitorctl"


class TestCommands(TestCase):
    def test_should_continue_when_user_accepts(self):
        # given
        out = StringIO()
        cmd = taskmonitorctl.Command(out, out)
        # when/then
        with patch(MODULE_PATH + ".my_input") as m:
            m.return_value = "y"
            cmd._user_confirmed("question")

    def test_should_stop_when_user_does_not_accept(self):
        # given
        out = StringIO()
        cmd = taskmonitorctl.Command(out, out)
        # when/then
        with self.assertRaises(SystemExit):
            with patch(MODULE_PATH + ".my_input") as m:
                m.return_value = "n"
                cmd._user_confirmed("question")

    def test_should_inspect_filled_logs(self):
        # given
        TaskLogFactory()
        out = StringIO()
        # when
        call_command("taskmonitorctl", "inspect", "logs", stdout=out)

    def test_should_inspect_empty_logs(self):
        # given
        out = StringIO()
        # when
        call_command("taskmonitorctl", "inspect", "logs", stdout=out)

    def test_should_inspect_empty_queue(self):
        # given
        out = StringIO()
        # when
        with patch(MODULE_PATH + ".celery_queues", spec=True) as m:
            m.queue_length.return_value = 0
            m._fetch_task_from_all_queues.return_value = []
            call_command("taskmonitorctl", "inspect", "queue", stdout=out)

    def test_should_inspect_filled_queue(self):
        # given
        out = StringIO()
        task = QueuedTaskShort.from_dict(QueuedTaskRawFactory())
        # when
        with patch(MODULE_PATH + ".celery_queues", spec=True) as m:
            m.queue_length.return_value = 1
            m._fetch_task_from_all_queues.return_value = [task]
            call_command("taskmonitorctl", "inspect", "queue", stdout=out)

    def test_should_inspect_settings(self):
        # given
        out = StringIO()
        # when
        call_command("taskmonitorctl", "inspect", "settings", stdout=out)

    @patch(MODULE_PATH + ".celery_queues", spec=True)
    def test_should_purge_empty_queue(self, mock_celery_queues):
        # given
        mock_celery_queues.queue_length.return_value = 0
        out = StringIO()
        # when
        with self.assertRaises(SystemExit):
            call_command("taskmonitorctl", "purge", "queue", "--all", stdout=out)
        # then
        self.assertFalse(mock_celery_queues.clear_tasks.called)

    @patch(MODULE_PATH + ".celery_queues", spec=True)
    def test_should_purge_filled_queue_all(self, mock_celery_queues):
        # given
        mock_celery_queues.queue_length.return_value = 1
        out = StringIO()
        # when
        with patch(MODULE_PATH + ".Command._user_confirmed", spec=True) as m:
            m.return_value = None
            call_command("taskmonitorctl", "purge", "queue", "--all", stdout=out)
        # then
        self.assertTrue(mock_celery_queues.clear_tasks.called)

    @patch(MODULE_PATH + ".celery_queues", spec=True)
    def test_should_purge_by_task_name(self, mock_celery_queues):
        # given
        mock_celery_queues.queue_length.return_value = 1
        mock_celery_queues.delete_task_by_name.return_value = 1
        out = StringIO()
        # when
        with patch(MODULE_PATH + ".Command._user_confirmed", spec=True) as m:
            m.return_value = None
            call_command(
                "taskmonitorctl", "purge", "queue", "--task-name", "alpha", stdout=out
            )
        # then
        self.assertTrue(mock_celery_queues.delete_task_by_name.called)

    @patch(MODULE_PATH + ".celery_queues", spec=True)
    def test_should_purge_by_task_id(self, mock_celery_queues):
        # given
        mock_celery_queues.queue_length.return_value = 1
        mock_celery_queues.delete_task_by_id.return_value = 1
        out = StringIO()
        # when
        with patch(MODULE_PATH + ".Command._user_confirmed", spec=True) as m:
            m.return_value = None
            call_command(
                "taskmonitorctl", "purge", "queue", "--task-id", "my-id", stdout=out
            )
        # then
        self.assertTrue(mock_celery_queues.delete_task_by_id.called)

    @patch(MODULE_PATH + ".celery_queues", spec=True)
    def test_should_purge_by_app_name(self, mock_celery_queues):
        # given
        mock_celery_queues.queue_length.return_value = 1
        mock_celery_queues.delete_task_by_app_name.return_value = 1
        out = StringIO()
        # when
        with patch(MODULE_PATH + ".Command._user_confirmed", spec=True) as m:
            m.return_value = None
            call_command(
                "taskmonitorctl", "purge", "queue", "--app-name", "alpha", stdout=out
            )
        # then
        self.assertTrue(mock_celery_queues.delete_task_by_app_name.called)

    @patch(MODULE_PATH + ".cached_reports", spec=True)
    def test_should_purge_logs_filled(self, mock_cached_reports):
        # given
        TaskLogFactory()
        out = StringIO()
        # when
        with patch(MODULE_PATH + ".Command._user_confirmed", spec=True) as m:
            m.return_value = None
            call_command("taskmonitorctl", "purge", "logs", "--all", stdout=out)
        # then
        self.assertEqual(TaskLog.objects.count(), 0)
        self.assertTrue(mock_cached_reports.refresh_cache.called)

    @patch(MODULE_PATH + ".cached_reports", spec=True)
    def test_should_purge_logs_empty(self, mock_cached_reports):
        # given
        out = StringIO()
        # when/then
        with self.assertRaises(SystemExit):
            with patch(MODULE_PATH + ".Command._user_confirmed", spec=True) as m:
                m.return_value = None
                call_command("taskmonitorctl", "purge", "logs", "--all", stdout=out)



# Generated by Django 4.0.10 on 2023-12-04 02:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("taskmonitor", "0007_alter_tasklog_exception"),
    ]

    operations = [
        migrations.CreateModel(
            name="TaskStatistic",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("app", models.CharField(max_length=255)),
                ("runtime_min", models.FloatField()),
                ("runtime_avg", models.FloatField()),
                ("runtime_max", models.FloatField()),
                ("runtime_total", models.FloatField()),
                ("runs_succeeded", models.IntegerField()),
                ("runs_failed", models.IntegerField()),
                ("runs_retried", models.IntegerField()),
                ("runs_total", models.IntegerField()),
            ],
            options={
                "managed": False,
            },
        ),
    ]



# flake8: noqa
"""Script for creating generated task logs for testing."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
from taskmonitor.core import cached_reports
from taskmonitor.models import TaskLog
from taskmonitor.tests.factories import TaskLogFactory

MAX_ENTRIES = 500_000
MAX_BATCH = 10_000

assert MAX_BATCH <= MAX_ENTRIES
assert MAX_ENTRIES % MAX_BATCH == 0
print(f"Generating a total of {MAX_ENTRIES:,} task logs...")
max_runs = MAX_ENTRIES // MAX_BATCH
for run in range(max_runs):
    print(f"Generating {MAX_BATCH:,} logs - {run + 1:,} / {max_runs:,}")
    objs = TaskLogFactory.build_batch(size=MAX_BATCH)
    TaskLog.objects.bulk_create(objs, batch_size=500, ignore_conflicts=True)
cached_reports.refresh_cache()
print("DONE!")



from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from taskmonitor.core import task_logs
from taskmonitor.models import TaskLog

from ..factories import SenderStub, TaskLogFactory

MODULE_PATH = "taskmonitor.core.task_logs"


class TestRunWhenEnabledDecorator(NoSocketsTestCase):
    @patch(MODULE_PATH + ".TASKMONITOR_ENABLED", True)
    def test_should_pass_when_enabled(self):
        # given
        @task_logs.run_when_enabled
        def dummy(x, y, z):
            nonlocal my_var
            my_var = 1

        my_var = 0

        # when
        dummy(1, 2, 3)
        # then
        self.assertEqual(my_var, 1)

    @patch(MODULE_PATH + ".TASKMONITOR_ENABLED", False)
    def test_should_not_pass_when_disabled(self):
        # given
        @task_logs.run_when_enabled
        def dummy(x, y, z):
            nonlocal my_var
            my_var = 1

        my_var = 0

        # when
        dummy(1, 2, 3)
        # then
        self.assertEqual(my_var, 0)


@patch(MODULE_PATH + ".celery_queues.queue_length_cached", lambda: 0)
@patch(MODULE_PATH + ".task_records.fetch", spec=True)
@patch(MODULE_PATH + "._run_housekeeping_if_stale", spec=True)
class TestTaskFailureHandler(NoSocketsTestCase):
    def test_should_create_log_for_complete_task(
        self, mock_run_housekeeping_if_stale, mock_task_records_fetch
    ):
        # given
        mock_task_records_fetch.return_value = now()
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        sender = SenderStub.create_from_obj(expected)
        # when
        task_logs.task_failure_handler_2(
            sender=sender, task_id=str(expected.task_id), exception=None
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )
        self.assertTrue(mock_run_housekeeping_if_stale.called)

    def test_should_create_log_for_task_without_delivery_info(
        self, mock_run_housekeeping_if_stale, mock_task_records_fetch
    ):
        # given
        mock_task_records_fetch.return_value = now()
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        sender = SenderStub.create_from_obj(expected)
        sender.request.delivery_info = None
        # when
        task_logs.task_failure_handler_2(
            sender=sender, task_id=str(expected.task_id), exception=None
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )
        self.assertTrue(mock_run_housekeeping_if_stale.called)


@patch(MODULE_PATH + ".celery_queues.queue_length_cached", lambda: 0)
@patch(MODULE_PATH + ".task_records.fetch", spec=True)
@patch(MODULE_PATH + "._run_housekeeping_if_stale", spec=True)
class TestTaskInternalErrorHandler(NoSocketsTestCase):
    def test_should_create_log_for_complete_task(
        self, mock_run_housekeeping_if_stale, mock_task_records_fetch
    ):
        # given
        mock_task_records_fetch.return_value = now()
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        sender = SenderStub.create_from_obj(expected)
        # when
        task_logs.task_internal_error_handler_2(
            sender=sender,
            task_id=str(expected.task_id),
            request=sender.request.asdict(),
            exception=None,
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )
        self.assertTrue(mock_run_housekeeping_if_stale.called)

    def test_should_create_log_for_task_without_delivery_info_1(
        self, mock_run_housekeeping_if_stale, mock_task_records_fetch
    ):
        # given
        mock_task_records_fetch.return_value = now()
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        sender = SenderStub.create_from_obj(expected)
        sender.request.delivery_info = None
        # when
        task_logs.task_internal_error_handler_2(
            sender=sender,
            task_id=str(expected.task_id),
            request=sender.request.asdict(),
            exception=None,
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )
        self.assertTrue(mock_run_housekeeping_if_stale.called)

    def test_should_create_log_for_task_without_delivery_info_2(
        self, mock_run_housekeeping_if_stale, mock_task_records_fetch
    ):
        # given
        mock_task_records_fetch.return_value = now()
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        sender = SenderStub.create_from_obj(expected)
        request = sender.request.asdict()
        del request["delivery_info"]
        # when
        task_logs.task_internal_error_handler_2(
            sender=sender,
            task_id=str(expected.task_id),
            request=request,
            exception=None,
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )
        self.assertTrue(mock_run_housekeeping_if_stale.called)



import datetime as dt
import json
from typing import List
from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from taskmonitor.core.celery_queues import QueuedTaskShort
from taskmonitor.models import QueuedTask, TaskLog

from .factories import QueuedTaskRawFactory, TaskLogFactory
from .fake_exceptions import make_fake_exception

MODELS_PATH = "taskmonitor.models"
MANAGERS_PATH = "taskmonitor.managers"


class TestTaskLog(TestCase):
    def test_should_convert_to_json(self):
        # given
        obj = TaskLogFactory()
        # when
        data = obj.asjson()
        # then
        obj_2 = json.loads(data)
        self.assertEqual(obj_2["task_name"], obj.task_name)
        self.assertEqual(obj_2["task_id"], str(obj.task_id))


class TestTaskLogQuerySet(TestCase):
    def test_should_return_oldest_date(self):
        # given
        log_1 = TaskLogFactory()
        log_2 = TaskLogFactory(received=log_1.received - dt.timedelta(hours=1))
        # when
        result = TaskLog.objects.oldest_date()
        # then
        self.assertEqual(result, log_2.timestamp)

    def test_should_return_none_when_no_logs_for_oldest_data(self):
        # when
        result = TaskLog.objects.oldest_date()
        # then
        self.assertIsNone(result)

    def test_should_return_newest_date(self):
        # given
        log_1 = TaskLogFactory()
        TaskLogFactory(received=log_1.received - dt.timedelta(hours=1))
        # when
        result = TaskLog.objects.newest_date()
        # then
        self.assertEqual(result, log_1.timestamp)

    def test_should_return_none_when_no_logs_for_newest_data(self):
        # when
        result = TaskLog.objects.newest_date()
        # then
        self.assertIsNone(result)

    def test_should_return_stale_logs(self):
        # given
        current_dt = timezone.now()
        TaskLogFactory(timestamp=current_dt)
        log_2 = TaskLogFactory(timestamp=current_dt - dt.timedelta(hours=2))
        log_3 = TaskLogFactory(timestamp=current_dt - dt.timedelta(hours=3))
        # when
        result = TaskLog.objects.filter_stale_logs(max_hours=1)
        # then
        pks = set(result.values_list("pk", flat=True))
        self.assertSetEqual(pks, {log_2.pk, log_3.pk})


class TestManagerCreateFromTask(TestCase):
    def test_should_create_from_succeeded_task(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, current_queue_length=42
        )
        # when
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = expected.timestamp
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
                result=expected.result,
                current_queue_length=42,
            )
        # then
        self._assert_equal_objs(expected, result)
        self.assertFalse(result.traceback)

    def test_should_create_from_failed_task(self):
        # given
        exception = make_fake_exception()
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="FakeException"
        )
        # when
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = expected.timestamp
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
                result=expected.result,
                exception=exception,
            )
        # then
        self._assert_equal_objs(expected, result)
        self.assertIn("Traceback", result.traceback)

    def test_should_truncate_args(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, args=[1, [1, 2], 3]
        )
        # when
        with patch(MANAGERS_PATH + ".TASKMONITOR_TRUNCATE_NESTED_DATA", True):
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
            )
        # then
        self.assertListEqual(result.args, [1, [], 3])

    def test_should_not_truncate_args(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, args=[1, [1, 2], 3]
        )
        # when
        with patch(MANAGERS_PATH + ".TASKMONITOR_TRUNCATE_NESTED_DATA", False):
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
            )
        # then
        self.assertListEqual(result.args, [1, [1, 2], 3])

    def test_should_truncate_kwargs(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, kwargs={"b": 2, "a": {"aa": 1}}
        )
        # when
        with patch(MANAGERS_PATH + ".TASKMONITOR_TRUNCATE_NESTED_DATA", True):
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
            )
        # then
        self.assertDictEqual(result.kwargs, {"a": {}, "b": 2})

    def test_should_not_truncate_kwargs(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, kwargs={"a": {"aa": 1}}
        )
        # when
        with patch(MANAGERS_PATH + ".TASKMONITOR_TRUNCATE_NESTED_DATA", False):
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
            )
        # then
        self.assertDictEqual(result.kwargs, {"a": {"aa": 1}})

    def test_should_truncate_result(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, result=[1, [1, 2], 3]
        )
        # when
        with patch(MANAGERS_PATH + ".TASKMONITOR_TRUNCATE_NESTED_DATA", True):
            obj = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
                result=expected.result,
            )
        # then
        self.assertListEqual(obj.result, [1, [], 3])

    def test_should_not_truncate_result(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.SUCCESS, result=[1, [1, 2], 3]
        )
        # when
        with patch(MANAGERS_PATH + ".TASKMONITOR_TRUNCATE_NESTED_DATA", False):
            obj = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=expected.kwargs,
                result=expected.result,
            )
        # then
        self.assertListEqual(obj.result, [1, [1, 2], 3])

    def test_should_handle_args_is_none(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback="", args=[]
        )
        # when
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = expected.timestamp
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=None,
                kwargs=expected.kwargs,
                result=expected.result,
            )
        # then
        self._assert_equal_objs(expected, result)

    def test_should_handle_kwargs_is_none(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback="", kwargs={}
        )
        # when
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = expected.timestamp
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
                args=expected.args,
                kwargs=None,
                result=expected.result,
            )
        # then
        self._assert_equal_objs(expected, result)

    def _assert_equal_objs(self, expected, result):
        field_names = {
            field.name
            for field in TaskLog._meta.fields
            if field.name not in {"id", "current_queue_length", "traceback"}
        }
        for field_name in field_names:
            with self.subTest(field_name=field_name):
                self.assertEqual(
                    getattr(expected, field_name), getattr(result, field_name)
                )


class TestCalcThroughput(TestCase):
    def test_should_calc_max(self):
        # given
        start = timezone.now().replace(second=0)
        TaskLogFactory(timestamp=start)
        TaskLogFactory(timestamp=start + dt.timedelta(minutes=0, seconds=2))
        TaskLogFactory(timestamp=start + dt.timedelta(minutes=0, seconds=3))
        TaskLogFactory(timestamp=start + dt.timedelta(minutes=1, seconds=1))
        # when
        self.assertEqual(TaskLog.objects.all().max_throughput(), 3)

    def test_should_calc_avg(self):
        # given
        start = timezone.now().replace(second=0)
        TaskLogFactory(timestamp=start)
        TaskLogFactory(timestamp=start + dt.timedelta(minutes=0, seconds=2))
        TaskLogFactory(timestamp=start + dt.timedelta(minutes=0, seconds=3))
        TaskLogFactory(timestamp=start + dt.timedelta(minutes=1, seconds=1))
        # when
        self.assertEqual(TaskLog.objects.all().avg_throughput(), 2)


class TestQueuedTask(TestCase):
    def test_should_create_objects_from_dict(self):
        # given
        queued_task_raw = QueuedTaskRawFactory()
        # when
        obj = QueuedTask.from_dict(queued_task_raw, 99)
        # then
        headers = queued_task_raw["headers"]
        self.assertEqual(obj.id, headers["id"])
        self.assertEqual(
            obj.name,
            headers["task"],
        )
        properties = queued_task_raw["properties"]
        self.assertEqual(
            obj.priority,
            properties["priority"],
        )
        self.assertEqual(obj.position, 99)

    def test_should_raise_error_when_dict_incomplete(self):
        # given
        queued_task_raw = {}
        # when
        with self.assertRaises(ValueError):
            QueuedTask.from_dict(queued_task_raw, 9)


def make_dto_list(objs) -> List[QueuedTaskShort]:
    """Convert list of raw tasks into task DTOs."""
    return [QueuedTaskShort.from_dict(obj) for obj in objs]


class TestQueuedTaskManager(TestCase):
    @patch("taskmonitor.managers.celery_queues", spec=True)
    def test_get_queryset_should_fetch_from_celery_queues(self, mock_celery_queues):
        # given
        queued_task_raw = QueuedTaskRawFactory()
        mock_celery_queues.queue_length.return_value = 2
        mock_celery_queues.fetch_tasks.return_value = make_dto_list(
            [queued_task_raw, QueuedTaskRawFactory()]
        )
        # when
        qs = QueuedTask.objects.all()
        # then
        self.assertEqual(len(qs), 2)
        self.assertEqual(qs[0].id, queued_task_raw["headers"]["id"])

    def test_should_support_all(self):
        # given
        tasks = make_dto_list([QueuedTaskRawFactory(), QueuedTaskRawFactory()])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when/then
        result = qs.all()
        self.assertEqual(qs, result)

    def test_should_support_count(self):
        # given
        tasks = make_dto_list([QueuedTaskRawFactory(), QueuedTaskRawFactory()])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when/then
        self.assertEqual(qs.count(), 2)

    def test_should_support_get(self):
        # given
        queued_task_raw = QueuedTaskRawFactory()
        tasks = make_dto_list(
            [
                QueuedTaskRawFactory(),
                QueuedTaskRawFactory(),
                QueuedTaskRawFactory(),
                queued_task_raw,
                QueuedTaskRawFactory(),
                QueuedTaskRawFactory(),
            ]
        )
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        obj = qs.get(id=queued_task_raw["headers"]["id"])
        # then
        self.assertEqual(obj.name, queued_task_raw["headers"]["task"])

    def test_should_support_first(self):
        # given
        queued_task_raw = QueuedTaskRawFactory()
        tasks = make_dto_list(
            [queued_task_raw, QueuedTaskRawFactory(), QueuedTaskRawFactory()]
        )
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        obj = qs.first()
        # then
        self.assertEqual(obj.name, queued_task_raw["headers"]["task"])

    def test_should_support_order_by_asc(self):
        # given
        self.maxDiff = None
        raw_1 = QueuedTaskRawFactory(headers__task="bravo")
        raw_2 = QueuedTaskRawFactory(headers__task="alpha")
        tasks = make_dto_list([raw_1, raw_2])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.order_by("name")
        # then
        task_ids = [obj.id for obj in result]
        self.assertEqual(task_ids, raw_task_ids([raw_2, raw_1]))

    def test_should_support_order_by_desc(self):
        # given
        self.maxDiff = None
        raw_1 = QueuedTaskRawFactory(headers__task="bravo")
        raw_2 = QueuedTaskRawFactory(headers__task="alpha")
        tasks = make_dto_list([raw_2, raw_1])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.order_by("-name")
        # then
        task_ids = [obj.id for obj in result]
        self.assertEqual(task_ids, raw_task_ids([raw_1, raw_2]))

    def test_order_by_multi(self):
        # given
        self.maxDiff = None
        raw_1 = QueuedTaskRawFactory(headers__task="bravo", properties__priority=7)
        raw_2 = QueuedTaskRawFactory(headers__task="bravo", properties__priority=1)
        raw_3 = QueuedTaskRawFactory(headers__task="alpha", properties__priority=4)
        tasks = make_dto_list([raw_1, raw_2, raw_3])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.order_by("name", "priority")
        # then
        task_ids = [obj.id for obj in result]
        self.assertEqual(task_ids, raw_task_ids([raw_3, raw_2, raw_1]))

    def test_should_support_filter(self):
        # given
        raw_1 = QueuedTaskRawFactory(headers__task="bravo")
        raw_2 = QueuedTaskRawFactory(headers__task="bravo")
        raw_3 = QueuedTaskRawFactory(headers__task="alpha")
        tasks = make_dto_list([raw_1, raw_2, raw_3])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.filter(name="alpha")
        # then
        task_ids = [obj.id for obj in result]
        self.assertEqual(task_ids, raw_task_ids([raw_3]))

    def test_should_raise_error_when_filter_with_unsupported_params(self):
        # given
        raw_1 = QueuedTaskRawFactory(headers__task="bravo")
        raw_2 = QueuedTaskRawFactory(headers__task="bravo")
        raw_3 = QueuedTaskRawFactory(headers__task="alpha")
        tasks = make_dto_list([raw_1, raw_2, raw_3])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.filter(name="alpha")
        # then
        task_ids = [obj.id for obj in result]
        self.assertEqual(task_ids, raw_task_ids([raw_3]))

    def test_should_support_values(self):
        # given
        raw_1 = QueuedTaskRawFactory(headers__task="alpha", properties__priority=5)
        raw_2 = QueuedTaskRawFactory(headers__task="bravo", properties__priority=5)
        tasks = make_dto_list([raw_1, raw_2])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.values("name", "priority")
        # then
        self.assertEqual(
            result, [{"name": "alpha", "priority": 5}, {"name": "bravo", "priority": 5}]
        )

    def test_should_support_values_list(self):
        # given
        raw_1 = QueuedTaskRawFactory(headers__task="alpha", properties__priority=5)
        raw_2 = QueuedTaskRawFactory(headers__task="bravo", properties__priority=5)
        tasks = make_dto_list([raw_1, raw_2])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.values_list("name", "priority")
        # then
        self.assertEqual(result, [("alpha", 5), ("bravo", 5)])

    def test_should_support_flat_values_list(self):
        # given
        raw_1 = QueuedTaskRawFactory(headers__task="alpha")
        raw_2 = QueuedTaskRawFactory(headers__task="bravo")
        tasks = make_dto_list([raw_1, raw_2])
        qs = QueuedTask.objects.from_dto_list(tasks)
        # when
        result = qs.values_list("name", flat=True)
        # then
        self.assertEqual(result, ["alpha", "bravo"])

    @patch("taskmonitor.managers.TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT", 100)
    @patch("taskmonitor.managers.celery_queues", spec=True)
    def test_get_queryset_should_return_empty_list_when_over_limit(
        self, mock_celery_queues
    ):
        # given
        mock_celery_queues.queue_length.return_value = 101
        # when
        qs = QueuedTask.objects.all()
        # then
        self.assertEqual(len(qs), 0)


def raw_task_ids(lst):
    return [obj["headers"]["id"] for obj in lst]



# Generated by Django 4.0.7 on 2022-09-26 20:47

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("taskmonitor", "0002_taskreport"),
    ]

    operations = [
        migrations.CreateModel(
            name="QueuedTask",
            fields=[
                ("app_name", models.CharField(max_length=255)),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("priority", models.PositiveIntegerField(default=None, null=True)),
                ("position", models.PositiveIntegerField()),
            ],
            options={
                "managed": False,
            },
        ),
    ]



import datetime as dt
from statistics import mean
from unittest.mock import MagicMock, patch

from pytz import utc

from django.test import TestCase
from django.utils.timezone import now

from taskmonitor.core import cached_reports
from taskmonitor.models import TaskLog

from ..factories import TaskLogFactory, fake_apps

MODULE_PATH = "taskmonitor.core.cached_reports"


@patch(MODULE_PATH + ".cache")
class TestCachedReports(TestCase):
    class DummyReport(cached_reports._CachedReport):
        def _calc_data(self):
            return "not cached"

    def test_should_create_obj(self, mock):
        # when
        obj = self.DummyReport()
        # then
        self.assertEqual(obj.name, "dummy_report")
        self.assertIn("tasklog", obj.changelist_url())

    def test_should_return_data_from_cache(self, mock_cache):
        # given
        mock_cache.get_or_set.return_value = "cached"
        report = self.DummyReport()
        # when
        data = report.data()
        # then
        self.assertEqual(data, "cached")

    def test_should_return_data_and_update_cache(self, mock_cache):
        # given
        report = self.DummyReport()
        # when
        data = report.data(use_cache=False)
        # then
        self.assertEqual(data, "not cached")
        self.assertTrue(mock_cache.set.called)


class TestCaseCachedReport(TestCase):
    def setUp(self) -> None:
        cached_reports.clear_cache()


class TestQueueLengthOverTime(TestCaseCachedReport):
    def test_should_create_queue_report(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        TaskLogFactory(
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=30,
        )
        TaskLogFactory(
            received=start_dt + dt.timedelta(seconds=5),
            started=start_dt + dt.timedelta(seconds=5),
            timestamp=start_dt + dt.timedelta(seconds=10),
            current_queue_length=10,
        )
        start_dt += dt.timedelta(minutes=5)
        TaskLogFactory(
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=50,
        )
        TaskLogFactory(
            received=start_dt + dt.timedelta(seconds=5),
            started=start_dt + dt.timedelta(seconds=5),
            timestamp=start_dt + dt.timedelta(seconds=10),
            current_queue_length=30,
        )
        report = cached_reports.QueueLengthOverTime()
        # when
        result = report._calc_data()
        # then
        series = result[0]
        self.assertEqual(series["name"], "length")
        data = series["data"]
        expected = [(1672574400000, 20), (1672574700000, 40)]
        self.assertEqual(data, expected)

    def test_should_work_with_null_values(self):
        # given
        TaskLogFactory(state=TaskLog.State.FAILURE)
        log = TaskLogFactory(state=TaskLog.State.SUCCESS, current_queue_length=None)
        self.assertIsNone(log.current_queue_length)
        report = cached_reports.QueueLengthOverTime()
        # when
        report._calc_data()

    def test_should_work_without_data(self):
        # given
        report = cached_reports.QueueLengthOverTime()
        # when
        report._calc_data()


class TestTruncateMinute(TestCaseCachedReport):
    @staticmethod
    def _make_qs(lst) -> MagicMock:
        m = MagicMock()
        m.values_list.return_value.iterator.return_value = lst
        return m

    def test_should_calc_mean(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        qs = self._make_qs(
            [
                (start_dt, 1),
                (start_dt, 3),
                (start_dt + dt.timedelta(minutes=2), 3),
                (start_dt + dt.timedelta(minutes=1), 3),
                (start_dt + dt.timedelta(minutes=1), 6),
            ]
        )
        # when
        result = cached_reports._CachedReport._truncate_minutes(mean, qs)
        # then
        expected = [(1672574400000, 2), (1672574460000, 4), (1672574520000, 3)]
        self.assertListEqual(result, expected)

    def test_should_calc_sum(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        qs = self._make_qs(
            [
                (start_dt, 1),
                (start_dt, 3),
                (start_dt + dt.timedelta(minutes=1), 3),
                (start_dt + dt.timedelta(minutes=1, seconds=5), 6),
                (start_dt + dt.timedelta(minutes=2), 3),
            ]
        )
        # when
        result = cached_reports._CachedReport._truncate_minutes(sum, qs)
        # then
        expected = [(1672574400000, 4), (1672574460000, 9), (1672574520000, 3)]
        self.assertListEqual(result, expected)

    def test_should_calc_sum_over_5_minutes(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        qs = self._make_qs(
            [
                (start_dt, 1),
                (start_dt, 3),
                (start_dt + dt.timedelta(minutes=1), 3),
                (start_dt + dt.timedelta(minutes=1, seconds=5), 6),
                (start_dt + dt.timedelta(minutes=6), 3),
            ]
        )
        # when
        result = cached_reports._CachedReport._truncate_minutes(sum, qs, 5)
        # then
        expected = [(1672574400000, 13), (1672574700000, 3)]
        self.assertListEqual(result, expected)


@patch(MODULE_PATH + ".report")
class TestTasksThroughputByApp(TestCaseCachedReport):
    def test_should_calc_with_no_data(self, mock_report):
        # given
        mock_report.return_value.data.return_value = None
        obj = cached_reports.TasksThroughputByApp()
        # when
        result = obj._calc_data()
        # then
        self.assertListEqual(result, [])


class TestExceptionsOverTime(TestCaseCachedReport):
    def test_should_create_report(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        TaskLogFactory(
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=30,
            exception="Bravo",
        )
        TaskLogFactory(
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=30,
            exception="Alpha",
        )
        start_dt += dt.timedelta(minutes=5)
        report = cached_reports.ExceptionsThroughput()
        # when
        result = report._calc_data()
        # then
        self.assertEqual(result[0]["name"], "Alpha")
        self.assertEqual(result[1]["name"], "Bravo")


class TestAppFailuresOverTime(TestCaseCachedReport):
    def test_should_create_report(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        app_1 = fake_apps[0]
        app_2 = fake_apps[1]
        TaskLogFactory(
            app_name=app_1,
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=30,
            exception="Bravo",
            state=TaskLog.State.FAILURE,
        )
        TaskLogFactory(
            app_name=app_1,
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=10),
            current_queue_length=30,
            exception="Alpha",
            state=TaskLog.State.FAILURE,
        )
        TaskLogFactory(
            app_name=app_2,
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=10),
            current_queue_length=30,
            state=TaskLog.State.SUCCESS,
        )
        report = cached_reports.AppFailuresOverTime()
        # when
        result = report._calc_data()
        # then
        apps = {obj["name"] for obj in result}
        self.assertSetEqual(apps, {app_1})


class TestCachedReport2(TestCaseCachedReport):
    """Testing a cached report without mocking the cache."""

    def test_should_create_report(self):
        # given
        start_dt = dt.datetime(2023, 1, 1, 12, 0, tzinfo=utc)
        TaskLogFactory(
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=30,
            exception="Bravo",
            state=TaskLog.State.FAILURE,
        )
        TaskLogFactory(
            received=start_dt,
            started=start_dt,
            timestamp=start_dt + dt.timedelta(seconds=5),
            current_queue_length=30,
            exception="Alpha",
            state=TaskLog.State.FAILURE,
        )
        start_dt += dt.timedelta(minutes=5)
        report = cached_reports.AppFailuresOverTime()
        # when
        result = report.data()
        # then
        self.assertTrue(result)
        self.assertAlmostEqual(
            report.last_update_at(), now(), delta=dt.timedelta(seconds=30)
        )
        self.assertLessEqual(report.last_update_at(), report.next_update_at())



from django.test import TestCase

from app_utils.testdata_factories import UserFactory

from taskmonitor.core import cached_reports, celery_queues
from taskmonitor.models import TaskStatistic
from taskmonitor.tests.factories import TaskLogFactory


class TestUI(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = UserFactory(is_staff=True, is_superuser=True)
        cls.log_1 = TaskLogFactory()
        cls.log_2 = TaskLogFactory()

    # FIXME
    # def test_should_show_reports_page(self):
    #     # given
    #     cached_reports.clear_cache()
    #     self.client.force_login(self.user)
    #     # when
    #     response = self.client.get("/taskmonitor/admin_taskmonitor_reports")
    #     # then
    #     self.assertEqual(response.status_code, 200)

    def test_should_show_tasklogs_changelist_page(self):
        # given
        cached_reports.clear_cache()
        self.client.force_login(self.user)
        # when
        response = self.client.get("/admin/taskmonitor/tasklog/")
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_show_tasklogs_change_page(self):
        # given
        cached_reports.clear_cache()
        self.client.force_login(self.user)
        # when
        response = self.client.get(
            f"/admin/taskmonitor/tasklog/{self.log_1.id}/change/"
        )
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_show_queuedtasks_page(self):
        # given
        celery_queues.tasks_cache.clear()
        self.client.force_login(self.user)
        # when
        response = self.client.get("/admin/taskmonitor/queuedtask/")
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_show_statistics_page(self):
        # given
        TaskStatistic.objects.clear_cache()
        self.client.force_login(self.user)
        # when
        response = self.client.get("/admin/taskmonitor/taskstatistic/")
        # then
        self.assertEqual(response.status_code, 200)



"""Models for Task Monitor."""

import json
import uuid

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

from .core.celery_queues import QueuedTaskShort
from .helpers import extract_app_name
from .managers import QueuedTaskManager, TaskLogManager, TaskStatisticManager

CHAR_FIELD_MAX_LENGTH = 255


class QueuedTask(models.Model):
    """A task that has been queued for later execution."""

    class Meta:
        managed = False

    app_name = models.CharField(max_length=CHAR_FIELD_MAX_LENGTH)
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(max_length=CHAR_FIELD_MAX_LENGTH)
    priority = models.PositiveIntegerField(null=True, default=None)
    position = models.PositiveIntegerField()

    objects = QueuedTaskManager()

    def __str__(self):
        return self.id

    @classmethod
    def from_dict(cls, obj: dict, position: int) -> "QueuedTask":
        """Create object from dictionary (DEPRECATED)."""
        if "headers" not in obj:
            raise ValueError("headers missing in obj")
        headers = obj["headers"]
        properties = obj["properties"] if "properties" in obj else {}
        task_name = headers["task"]
        return cls(
            app_name=extract_app_name(task_name),
            id=headers["id"],
            name=task_name,
            priority=properties.get("priority"),
            position=position,
        )

    @classmethod
    def from_dto(cls, obj: QueuedTaskShort, position: int) -> "QueuedTask":
        """Create model object from data transport object."""
        return cls(
            app_name=obj.app_name,
            id=obj.id,
            name=obj.name,
            priority=obj.priority,
            position=position,
        )


class TaskReport(models.Model):
    """Dummy model to fake a 'Reports' entry on the admin index page."""

    class Meta:
        managed = False
        verbose_name = "report"

    id = models.BigIntegerField(primary_key=True)


class TaskStatistic(models.Model):
    """Statistics for a task.

    This is a fake model, which we need to this data on the admin site.
    """

    class Meta:
        managed = False

    name = models.CharField(max_length=CHAR_FIELD_MAX_LENGTH)
    app = models.CharField(max_length=CHAR_FIELD_MAX_LENGTH)
    runtime_min = models.FloatField()
    runtime_avg = models.FloatField()
    runtime_max = models.FloatField()
    runtime_total = models.FloatField()
    runs_succeeded = models.IntegerField()
    runs_failed = models.IntegerField()
    runs_retried = models.IntegerField()
    runs_total = models.IntegerField()

    objects = TaskStatisticManager()

    def __str__(self):
        return self.name


class TaskLog(models.Model):
    """Log entry for an executed celery task."""

    class State(models.IntegerChoices):
        """A state of a task log."""

        SUCCESS = 1, "success"
        RETRY = 2, "retry"
        FAILURE = 3, "failure"

    app_name = models.CharField(
        max_length=CHAR_FIELD_MAX_LENGTH,
        db_index=True,
        help_text="Name of the app this task belongs to.",
    )
    args = models.JSONField(
        default=list,
        help_text=(
            "Positional arguments the task was called with. Nested items might be truncated to [] or {}."
        ),
    )
    current_queue_length = models.IntegerField(
        default=None,
        null=True,
        help_text="Length of the queue at the time this log was created.",
    )
    exception = models.CharField(
        max_length=CHAR_FIELD_MAX_LENGTH,
        db_index=True,
        help_text="Name of the raised exception if any.",
    )
    kwargs = models.JSONField(
        default=dict,
        help_text=(
            "Keyword arguments the task was called with. Nested items might be truncated to [] or {}."
        ),
    )
    parent_id = models.UUIDField(
        null=True, default=None, help_text="ID of the parent task if any."
    )
    priority = models.IntegerField(
        null=True,
        default=None,
        db_index=True,
        help_text="Priority this task was executed with.",
    )
    result = models.JSONField(default=None, null=True, help_text="Result of the task.")
    retries = models.IntegerField(help_text="Number of retries.")
    received = models.DateTimeField(
        null=True,
        default=None,
        help_text="When a task is received from the broker and is ready for execution.",
    )
    runtime = models.FloatField(
        null=True,
        default=None,
        db_index=True,
        help_text="Runtime of this task in seconds.",
    )
    started = models.DateTimeField(
        null=True,
        default=None,
        help_text="When the task execution started.",
    )
    state = models.IntegerField(
        choices=State.choices,
        db_index=True,
        help_text="Task's state when it was logged.",
    )
    task_id = models.UUIDField(
        default=uuid.uuid4, db_index=True, help_text="Unique ID of this task."
    )
    task_name = models.CharField(
        max_length=CHAR_FIELD_MAX_LENGTH, db_index=True, help_text="Name of this task."
    )

    timestamp = models.DateTimeField(
        db_index=True, help_text="Timestamp when this log was created."
    )
    traceback = models.TextField(
        help_text="Full stack trace if there was an exception."
    )

    objects = TaskLogManager()

    def __str__(self):
        return f"{self.task_name}:{self.pk}"

    def save(self, *args, **kwargs) -> None:
        if self.started:
            self.runtime = (self.timestamp - self.started).total_seconds()
        super().save(*args, **kwargs)

    @property
    def timestamp_formatted(self) -> str:
        """Timestamp as string in default format."""
        return self.timestamp.strftime("%Y-%m-%d %H:%M:%S,%f")[:-3]

    def asdict(self) -> dict:
        """Convert to representation as Python dict."""
        struct = {}
        for field in self._meta.fields:
            if field.choices:
                value = getattr(self, f"get_{field.name}_display")()
            else:
                value = getattr(self, field.name)
            # if callable(value):
            #     try:
            #         value = value() or ""
            #     except Exception:
            #         value = "Error retrieving value"
            if value is None:
                value = ""
            struct[field.name] = value
        return struct

    def asjson(self) -> str:
        """Convert to representation in JSON."""
        return json.dumps(
            self.asdict(), indent=4, sort_keys=True, cls=DjangoJSONEncoder
        )

    def astext(self) -> str:
        """Convert to text form that can be easily shared, e.g. on Discord chat."""
        text = "Task Log:"
        for key, value in self.asdict().items():
            text += "\n\n"
            if isinstance(value, (dict, list, tuple, set)) or key == "traceback":
                text += f"{key}:\n{value}"
            else:
                text += f"{key}: {value}"
        return text



# Generated by Django 4.0.10 on 2023-07-04 13:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("taskmonitor", "0006_truncate_exceptions"),
    ]

    operations = [
        migrations.AlterField(
            model_name="tasklog",
            name="exception",
            field=models.CharField(
                db_index=True,
                help_text="Name of the raised exception if any.",
                max_length=255,
            ),
        ),
    ]



import datetime as dt
import itertools
from dataclasses import asdict, dataclass
from random import choice, choices, randint, shuffle
from uuid import UUID

import factory
import factory.fuzzy
from factory.faker import faker

from django.utils import timezone

from taskmonitor.models import TaskLog

# generate fake apps and task names
faker = faker.Faker()
_fake_tasks = {}
for app_name in {faker.first_name().lower() for _ in range(24)}:
    _fake_tasks[app_name] = [
        app_name + ".tasks." + "_".join(faker.words(3)).lower()
        for _ in range(randint(3, 20))
    ]
_fake_tasks_all = list(
    itertools.chain(*[_fake_tasks[app_name] for app_name in _fake_tasks])
)
fake_apps = list(_fake_tasks.keys())

_fake_words = [obj.lower() for obj in faker.words(50)]
_fake_numbers = [randint(0, 1_000_000) for _ in range(100)]
_fake_args = _fake_words + _fake_numbers
shuffle(_fake_args)


class TaskLogFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TaskLog

    app_name = factory.fuzzy.FuzzyChoice(fake_apps)
    current_queue_length = factory.fuzzy.FuzzyInteger(0, 1_000)
    received = factory.fuzzy.FuzzyDateTime(
        timezone.now() - dt.timedelta(hours=23, minutes=59)
    )
    started = factory.LazyAttribute(
        lambda o: factory.fuzzy.FuzzyDateTime(
            o.received, end_dt=o.received + dt.timedelta(minutes=1)
        ).fuzz()
    )
    task_name = factory.LazyAttribute(lambda o: choice(_fake_tasks[o.app_name]))

    @factory.lazy_attribute
    def task_id(self):
        return UUID(faker.uuid4())

    @factory.lazy_attribute
    def args(self):
        return list(choices(_fake_args, k=randint(0, 10)))

    @factory.lazy_attribute
    def kwargs(self):
        keys = choices(_fake_words, k=randint(0, 20))
        return {key: randint(0, 1_000_000) for key in keys}

    @factory.lazy_attribute
    def runtime(self):
        return (self.timestamp - self.started).total_seconds()

    @factory.lazy_attribute
    def priority(self):
        return choices(
            population=[1, 2, 3, 4, 5, 6, 7, 8, 9],
            weights=[5, 5, 10, 20, 100, 20, 10, 5, 5],
        )[0]

    @factory.lazy_attribute
    def retries(self):
        return choices(
            population=[0, 1, 2, 3],
            weights=[80, 10, 5, 5],
        )[0]

    @factory.lazy_attribute
    def state(self):
        return choices(
            population=[
                TaskLog.State.SUCCESS,
                TaskLog.State.RETRY,
                TaskLog.State.FAILURE,
            ],
            weights=[80, 15, 5],
        )[0]

    @factory.lazy_attribute
    def exception(self):
        if self.state == TaskLog.State.SUCCESS:
            return ""
        return faker.word().title()

    @factory.lazy_attribute
    def traceback(self):
        if self.state == TaskLog.State.SUCCESS:
            return ""
        return faker.paragraph()

    @factory.lazy_attribute
    def timestamp(self):
        max_duration = choices(
            population=[0.5, 1, 10, 30, 120],
            weights=[75, 10, 5, 5, 5],
        )[0]
        start_dt = self.started + dt.timedelta(seconds=0.1)
        return factory.fuzzy.FuzzyDateTime(
            start_dt=start_dt,
            end_dt=start_dt + dt.timedelta(seconds=max_duration),
        ).fuzz()

    @factory.lazy_attribute
    def result(self):
        if self.state == TaskLog.State.SUCCESS:
            return choice(_fake_args)
        return None


@dataclass
class ContextStub:
    id: str
    retries: int
    delivery_info: dict
    args: list
    kwargs: dict
    parent_id: str = None

    def asdict(self) -> dict:
        return asdict(self)

    @classmethod
    def create_from_obj(cls, obj: TaskLog):
        return cls(
            parent_id=obj.parent_id,
            retries=obj.retries,
            id=str(obj.task_id),
            delivery_info={
                "is_eager": False,
                "exchange": None,
                "routing_key": None,
                "priority": obj.priority,
            },
            args=obj.args,
            kwargs=obj.kwargs,
        )


@dataclass
class SenderStub:
    name: str
    request: ContextStub
    priority: int

    @classmethod
    def create_from_obj(cls, obj: TaskLog):
        request = ContextStub.create_from_obj(obj)
        return cls(name=obj.task_name, request=request, priority=5)


class QueuedTaskRawFactory(factory.DictFactory):
    body = (
        "W1tdLCB7fSwgeyJjYWxsYmFja3MiOiBudWxsLCAiZXJyYmFja3MiOi"
        "BudWxsLCAiY2hhaW4iOiBudWxsLCAiY2hvcmQiOiBudWxsfV0="
    )
    content_encoding = "utf-8"
    content_type = "application/json"
    headers = factory.Dict(
        {
            "lang": "py",
            "task": factory.LazyAttribute(lambda o: choice(_fake_tasks_all)),
            "id": factory.Faker("uuid4"),
            "retries": 0,
            "origin": "dummy@QueuedTaskRawFactory",
            "root_id": None,
            "parent_id": factory.LazyAttribute(lambda o: o.id),
        }
    )
    properties = factory.Dict({"priority": factory.fuzzy.FuzzyInteger(0, 9)})

    class Meta:
        rename = {
            "content_encoding": "content-encoding",
            "content_type": "content-type",
        }

    #     {
    #         "body": "W1tdLCB7fSwgeyJjYWxsYmFja3MiOiBudWxsLCAiZXJyYmFja3MiOiBudWxsLCAiY2hhaW4iOiBudWxsLCAiY2hvcmQiOiBudWxsfV0=",
    #         "content-encoding": "utf-8",
    #         "content-type": "application/json",
    #         "headers": {
    #             "lang": "py",
    #             "task": "taskmonitor.tasks.run_housekeeping",
    #             "id": "7b2416f0-1ff5-4c77-8403-8aafac05b298",
    #             "shadow": None,
    #             "eta": None,
    #             "expires": None,
    #             "group": None,
    #             "group_index": None,
    #             "retries": 0,
    #             "timelimit": [None, None],
    #             "root_id": "7b2416f0-1ff5-4c77-8403-8aafac05b298",
    #             "parent_id": None,
    #             "argsrepr": "()",
    #             "kwargsrepr": "{}",
    #             "origin": "gen9996@bji74-PC",
    #             "ignore_result": False,
    #         },
    #         "properties": {
    #             "correlation_id": "7b2416f0-1ff5-4c77-8403-8aafac05b298",
    #             "reply_to": "853ed4b4-2c52-32cd-be57-4086564ac31e",
    #             "delivery_mode": 2,
    #             "delivery_info": {"exchange": "", "routing_key": "celery"},
    #             "priority": 0,
    #             "body_encoding": "base64",
    #             "delivery_tag": "84f0b41d-c8fc-4dab-88e8-a89afddb1deb",
    #         },
    #     }
    # ]



"""Truncate exception field in preparation of changing it from text to char."""

from django.db import migrations


def forwards(apps, schema_editor):
    """Apply a forwards migration."""

    TaskLog = apps.get_model("taskmonitor", "TaskLog")

    for obj in TaskLog.objects.exclude(exception=""):
        parts = obj.exception.split("(")
        obj.exception = parts[0][:254]
        obj.save()


class Migration(migrations.Migration):
    dependencies = [
        ("taskmonitor", "0005_tasklog_current_queue_length_alter_tasklog_priority"),
    ]

    operations = [migrations.RunPython(forwards, migrations.RunPython.noop)]



"""Container for caching the reports data used in views."""

# pylint: disable = missing-class-docstring

import datetime as dt
import inspect
import re
import sys
from collections import defaultdict
from statistics import mean
from typing import Callable, Iterable, List, Optional, Tuple

from django.core.cache import cache
from django.db.models import Avg, Count, F, Max, Sum, Value
from django.db.models.functions import Concat, TruncMinute
from django.urls import reverse
from django.utils import timezone

from taskmonitor.app_settings import (
    TASKMONITOR_HOUSEKEEPING_FREQUENCY,
    TASKMONITOR_REPORTS_MAX_AGE,
    TASKMONITOR_REPORTS_MAX_TOP,
)
from taskmonitor.models import TaskLog

CACHE_KEY = "taskmonitor_reports_data"
MAX_APPS_COUNT = 14
APP_NAME_OTHERS = "Others"


class _CachedReport:
    """Base class for a cached report."""

    is_included = True  # whether a report is included in the main group

    def __init__(self) -> None:
        self.name = self._to_snake_case(self.__class__.__name__)

    @property
    def cache_key(self):
        """Return cache key."""
        return f"{CACHE_KEY}_{self.name}"

    @property
    def timeout(self):
        """Timeout in seconds."""
        return TASKMONITOR_REPORTS_MAX_AGE * 60

    @property
    def now(self) -> dt.datetime:
        """Return current date and time."""
        return timezone.now()

    @staticmethod
    def changelist_url() -> str:
        """Return changelist URL."""
        return reverse("admin:taskmonitor_tasklog_changelist")

    def data(self, use_cache=True):
        """Return data for this report."""
        if use_cache:
            return cache.get_or_set(
                self.cache_key, self._calc_data, timeout=self.timeout
            )
        calculated_data = self._calc_data()
        cache.set(self.cache_key, calculated_data, timeout=self.timeout)
        return calculated_data

    def refresh_cache(self) -> None:
        """Refresh the cache."""
        cache.set(self.cache_key, self._calc_data(), timeout=self.timeout)

    def clear_cache(self) -> None:
        """Clear the cache."""
        cache.delete(self.cache_key)

    def last_update_at(self) -> Optional[dt.datetime]:
        """When the cache was last updated or None if there is no cache."""
        ttl = self._ttl()
        if not ttl:
            return None
        return timezone.now() - dt.timedelta(seconds=max(0, self.timeout - ttl))

    def next_update_at(self) -> Optional[dt.datetime]:
        """When the cache will be updated next (earliest) or None if no cache."""
        ttl = self._ttl()
        if not ttl:
            return None
        duration = TASKMONITOR_HOUSEKEEPING_FREQUENCY * 60 / 2 + ttl
        return timezone.now() + dt.timedelta(seconds=duration)

    def _ttl(self):
        return cache.ttl(self.cache_key)

    def _calc_data(self):
        """Calculate data."""
        raise NotImplementedError()

    @staticmethod
    def _truncate_minutes(
        func: Callable, qs: Iterable, minutes: int = 1
    ) -> List[Tuple[int, int]]:
        """Truncate data to one aggregated value over a span of minutes.

        Result will be sorted ascending by datetime.

        Args:
            - func: function which takes a list of values and returns one value
            - values: iterable of values in the format: [{"x": 1, "y": 2}, ...]
            - minutes: number of minutes to apply the function over.
            Must be a divider of 60.

        This function is necessary, because tests show that the ORM grouping
        with TruncMinute is not reliable.
        """

        def _func_or_zero(func, lst) -> int:
            return func(lst) if lst else 0

        if 60 % minutes > 0:
            raise ValueError("minutes must be a divider of 60.")
        data_raw = defaultdict(list)
        for x, y in qs.values_list("x", "y").iterator():
            new_minutes = x.minute // minutes * minutes
            new_x = x.replace(minute=new_minutes, second=0, microsecond=0)
            x_timestamp = int(new_x.timestamp() * 1000)
            data_raw[x_timestamp].append(y)
        data_raw = dict(sorted(data_raw.items()))
        my_data = [
            tuple([x, int(round(_func_or_zero(func, values), 0))])
            for x, values in data_raw.items()
        ]
        return my_data

    @staticmethod
    def _to_snake_case(name: str) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()

    @classmethod
    def report_classes(cls):
        """Return all known report classes."""
        return [
            obj
            for _, obj in inspect.getmembers(sys.modules[__name__], inspect.isclass)
            if issubclass(obj, cls) and obj is not cls
        ]


class AppsTopCumRuntime(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runtime"]:
            return None
        return list(
            TaskLog.objects.values(name=F("app_name"))
            .annotate(y=Sum("runtime"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?o=5&app_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class AppsTopRuns(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runs"]:
            return None
        return list(
            TaskLog.objects.values(name=F("app_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?app_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksBasics(_CachedReport):
    """Basic information about tasks used by many other reports."""

    def _calc_data(self):
        total_runs = TaskLog.objects.count()
        total_runtime = TaskLog.objects.aggregate(total_runtime=Sum("runtime"))[
            "total_runtime"
        ]
        oldest_date = TaskLog.objects.oldest_date()
        newest_date = TaskLog.objects.newest_date()
        return {
            "total_runs": total_runs,
            "total_runtime": total_runtime,
            "oldest_date": oldest_date,
            "newest_date": newest_date,
        }


class QueueLengthOverTime(_CachedReport):
    is_included = False

    def _calc_data(self):
        qs = (
            TaskLog.objects.exclude(current_queue_length__isnull=True)
            .order_by("timestamp")
            .annotate(x=TruncMinute("timestamp"))
            .values("x")
            .annotate(y=Avg("current_queue_length"))
        )
        my_data = self._truncate_minutes(mean, qs, 5)
        return [{"name": "length", "data": my_data}]


class TaskStatistics(_CachedReport):
    def _calc_data(self):
        qs = (
            TaskLog.objects.values("app_name", "task_name")
            .annotate(x=Count("pk"))
            .annotate(y=Avg("runtime"))
        )
        grouped_by_app = defaultdict(list)
        for obj in qs:
            point = {
                "x": obj["x"],
                "y": obj["y"],
                "task_name": obj["task_name"],
            }
            grouped_by_app[obj["app_name"]].append(point)
        my_data = [
            {"name": name, "data": data} for name, data in grouped_by_app.items()
        ]
        return my_data


class TaskRunsByState(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runs"]:
            return None
        return [
            {
                "name": state.label,
                "y": TaskLog.objects.filter(state=state.value).count(),
                "url": f"{self.changelist_url()}?state__exact={state}",
            }
            for state in TaskLog.State
        ]


class TaskRunsByApp(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runs"]:
            return None
        my_data = list(
            TaskLog.objects.values(name=F("app_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?app_name="), F("name"))
            )
            .order_by("-y")
        )
        if len(my_data) > MAX_APPS_COUNT:
            others_y = sum(
                (
                    app["y"]
                    for i, app in enumerate(my_data, start=1)
                    if i > MAX_APPS_COUNT
                )
            )
            return my_data[:MAX_APPS_COUNT] + [
                {"name": APP_NAME_OTHERS, "y": others_y, "url": "#"}
            ]
        return my_data


class TasksTopRuns(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runs"]:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?task_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopAvgRuntime(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runtime"]:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Avg("runtime"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?o=5&task_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopCumRuntime(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runtime"]:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Sum("runtime"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?o=5&task_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopMaxRuntime(_CachedReport):
    def _calc_data(self):
        if not report("tasks_basics").data()["total_runtime"]:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Max("runtime"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url()}?o=5&task_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopFailed(_CachedReport):
    def _calc_data(self):
        total_failed = TaskLog.objects.filter(state=TaskLog.State.FAILURE).count()
        if not total_failed:
            return None
        return list(
            TaskLog.objects.filter(state=TaskLog.State.FAILURE)
            .values(name=F("task_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(
                    Value(f"{self.changelist_url()}?state__exact=3&task_name="),
                    F("name"),
                )
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopRetried(_CachedReport):
    def _calc_data(self):
        total_retried = TaskLog.objects.filter(state=TaskLog.State.RETRY).count()
        if not total_retried:
            return None

        return list(
            TaskLog.objects.filter(state=TaskLog.State.RETRY)
            .values(name=F("task_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(
                    Value(f"{self.changelist_url()}?state__exact=2&task_name="),
                    F("name"),
                )
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksThroughput(_CachedReport):
    def _calc_data(self):
        tasklogs_not_failed = TaskLog.objects.filter(state=TaskLog.State.SUCCESS)
        tasks_throughput = []
        average_last_hours = {}
        for hours in [1, 3, 6, 12, 24]:
            average_last_hours[hours] = tasklogs_not_failed.filter(
                timestamp__gt=self.now - dt.timedelta(hours=hours)
            ).avg_throughput()
        for hours, y in average_last_hours.items():
            tasks_throughput.append({"name": f"Average last {hours} hours", "y": y})
        average_overall = tasklogs_not_failed.avg_throughput()
        tasks_throughput.append({"name": "Average overall", "y": average_overall})
        peak_overall = tasklogs_not_failed.max_throughput()
        tasks_throughput.append({"name": "Peak overall", "y": peak_overall})
        return tasks_throughput


class TasksThroughputByState(_CachedReport):
    is_included = False

    def _calc_data(self):
        series = []
        for state in TaskLog.State:
            qs = (
                TaskLog.objects.filter(state=state)
                .order_by("timestamp")
                .annotate(x=TruncMinute("timestamp"))
                .values("x")
                .annotate(y=Count("id"))
            )
            my_data = self._truncate_minutes(sum, qs, 5)
            series.append({"name": state.label, "data": my_data})
        return series


class TasksThroughputByApp(_CachedReport):
    is_included = False

    def _calc_data(self):
        series = []
        app_names = (
            TaskLog.objects.filter(state=TaskLog.State.SUCCESS)
            .values_list("app_name", flat=True)
            .distinct()
            .order_by("app_name")
        )
        for app_name in app_names:
            qs = (
                TaskLog.objects.filter(state=TaskLog.State.SUCCESS, app_name=app_name)
                .order_by("timestamp")
                .annotate(x=TruncMinute("timestamp"))
                .values("x")
                .annotate(y=Count("id"))
            )
            my_data = self._truncate_minutes(sum, qs, 5)
            series.append({"name": app_name, "data": my_data})
        return series


class ExceptionsThroughput(_CachedReport):
    is_included = False

    def _calc_data(self):
        series = []
        exceptions = (
            TaskLog.objects.exclude(exception="")
            .exclude(exception="Retry")
            .values_list("exception", flat=True)
            .distinct()
            .order_by("exception")
        )
        for exception in exceptions:
            app_qs = TaskLog.objects.filter(exception=exception)
            qs = (
                app_qs.order_by("timestamp")
                .annotate(x=TruncMinute("timestamp"))
                .values("x")
                .annotate(y=Count("id"))
            )
            my_data = self._truncate_minutes(sum, qs, 5)
            series.append({"name": exception, "data": my_data})
        return series


class AppFailuresOverTime(_CachedReport):
    is_included = False

    def _calc_data(self):
        series = []
        app_names = (
            TaskLog.objects.filter(state=TaskLog.State.FAILURE)
            .values_list("app_name", flat=True)
            .distinct()
            .order_by("app_name")
        )
        for app_name in app_names:
            qs = (
                TaskLog.objects.filter(app_name=app_name, state=TaskLog.State.FAILURE)
                .order_by("timestamp")
                .annotate(x=TruncMinute("timestamp"))
                .values("x")
                .annotate(y=Count("id"))
            )
            my_data = self._truncate_minutes(sum, qs, 5)
            series.append({"name": app_name, "data": my_data})
        return series


def refresh_cache() -> None:
    """Refresh the cache."""
    for my_report in reports():
        my_report.refresh_cache()


def clear_cache() -> None:
    """Clear cache for all reports."""
    for my_report in reports():
        my_report.clear_cache()


def data() -> dict:
    """Calculate the report data."""
    return {
        my_report.name: report_data(my_report.name)
        for my_report in reports()
        if my_report.is_included
    }


def report_data(report_name: str, use_cache: bool = True):
    """Data of an cached report."""
    return report(report_name).data(use_cache)


def reports() -> List[_CachedReport]:
    """List of all cached reports."""
    return _reports.values()


def report(report_name: str) -> _CachedReport:
    """Access report by name."""
    return _reports[report_name]


# Instantiation of all cached reports
_reports = {obj.name: obj for obj in [cls() for cls in _CachedReport.report_classes()]}



import datetime as dt
from unittest.mock import patch

from django.test import TestCase, override_settings
from django.utils import timezone

from taskmonitor import tasks
from taskmonitor.models import TaskLog

from .factories import TaskLogFactory

TASKS_PATH = "taskmonitor.tasks"


class TestDeleteStaleTasklogs(TestCase):
    @patch(TASKS_PATH + ".TaskLog", wraps=TaskLog)
    def test_should_delete_stale_entries_only(self, spy_TaskLog):
        # given
        current_dt = timezone.now()
        TaskLogFactory(timestamp=current_dt - dt.timedelta(hours=3, seconds=1))
        TaskLogFactory(timestamp=current_dt - dt.timedelta(hours=3, seconds=1))
        TaskLogFactory(timestamp=current_dt - dt.timedelta(hours=3, seconds=1))
        current_entry = TaskLogFactory(timestamp=current_dt)
        # when
        with patch(TASKS_PATH + ".TASKMONITOR_DELETE_STALE_BATCH_SIZE", 2), patch(
            TASKS_PATH + ".TASKMONITOR_DATA_MAX_AGE", 3
        ):
            tasks.delete_stale_tasklogs()
        # then
        self.assertEqual(TaskLog.objects.count(), 1)
        self.assertTrue(TaskLog.objects.filter(pk=current_entry.pk).exists())
        self.assertEqual(spy_TaskLog.objects.filter.call_count, 2)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestRefreshCachedData(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        TaskLogFactory()
        TaskLogFactory()

    @patch(TASKS_PATH + ".refresh_statistics_cache")
    @patch(TASKS_PATH + ".refresh_single_report_cache")
    def test_should_refresh_cached_data(
        self, mock_refresh_single_report_cache, mock_refresh_statistics_cache
    ):
        # when
        tasks.refresh_cached_data()
        # then
        self.assertEqual(mock_refresh_single_report_cache.apply_async.call_count, 18)
        self.assertTrue(mock_refresh_statistics_cache.apply_async.called)

    @patch(TASKS_PATH + ".cached_reports", spec=True)
    def test_should_refresh_reports_cache(self, mock_cached_report):
        # when
        tasks.refresh_single_report_cache("dummy")
        # then
        self.assertTrue(mock_cached_report.report.return_value.refresh_cache.called)

    @patch(TASKS_PATH + ".TaskStatistic")
    def test_should_refresh_statistics_cache(self, mock_TaskStatistic):
        # when
        tasks.refresh_statistics_cache()
        # then
        self.assertTrue(mock_TaskStatistic.objects.refresh_cache.called)



"""A global key/value store for task related data.

The django cache is used as data storage to enable persistence and thread safety.
"""

# pylint: disable = redefined-builtin

from typing import Any, Optional

from django.core.cache import cache

from taskmonitor.app_settings import TASKMONITOR_DATA_MAX_AGE

CACHE_KEY = "taskmonitor_records"


def set(task_id: str, key: str, value: Any):
    """Set a key/value for a task."""
    timeout = TASKMONITOR_DATA_MAX_AGE * 3600
    cache.set(_build_key(task_id, key), value, timeout=timeout)


def get(task_id: str, key: str) -> Optional[Any]:
    """Returns the value for a task and key."""
    return cache.get(_build_key(task_id, key))


def delete(task_id: str, key: str) -> None:
    """Remove a key for a task."""
    cache.delete(_build_key(task_id, key))


def fetch(task_id: str, key: str) -> Optional[Any]:
    """Fetch the value and removes it's key for task."""
    value = get(task_id, key)
    delete(task_id, key)
    return value


def _build_key(task_id, key: str) -> str:
    return f"{CACHE_KEY}_{task_id}_{key}"



"""API for working with celery task queue."""

import concurrent.futures
import functools
import itertools
import json
from collections import defaultdict
from typing import List

import redis

from django.conf import settings

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from taskmonitor import __title__
from taskmonitor.app_settings import TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT
from taskmonitor.helpers import memcached

from .tasks_cache import QueuedTasksCache, QueuedTaskShort

PRIORITY_SEP = "\x06\x16"
DEFAULT_PRIORITY_STEPS = range(10)
QUEUED_TASKS_CACHE_KEY = "queued-tasks-cache-key"

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def _redis_client():
    """Fetch the Redis client for the celery broker."""
    return redis.from_url(settings.BROKER_URL)


def default_queue_name() -> str:
    """Default name of celery queue in AA."""
    return getattr(settings, "CELERY_DEFAULT_QUEUE", "celery")


def _redis_queue_names(queue_name: str = None) -> List[str]:
    """List of all queue names on Redis incl. the dedicated queue names for each priority."""
    if not queue_name:
        queue_name = default_queue_name()
    names = [
        f"{queue_name}{PRIORITY_SEP}{priority}" for priority in DEFAULT_PRIORITY_STEPS
    ]
    names = [queue_name] + names
    return names


def queue_length() -> int:
    """Length of the celery queue."""
    r = _redis_client()
    return sum(r.llen(name) for name in _redis_queue_names())


@memcached(timeout=10)
def queue_length_cached() -> int:
    """Current queue length, but cached for a couple seconds to reduce load on Redis."""
    return queue_length()


def fetch_tasks() -> List[QueuedTaskShort]:
    """Fetch tasks from queues
    and return as ordered list with oldest task in first position.
    """
    data = tasks_cache.get()
    if data is None:
        logger.debug("Cache is stale. Fetching new tasks from queue.")
        tasks = _fetch_task_from_all_queues()
        tasks_cache.set(tasks)
    else:
        tasks = data.tasks
        logger.debug("Returning tasks from cache.")
    return tasks


def _fetch_task_from_all_queues() -> List[QueuedTaskShort]:
    """Coordinate fetching tasks from all priority queues."""
    _fetch_func = functools.partial(_fetch_tasks_from_redis, _redis_client())
    redis_queue_names = _redis_queue_names()
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=len(redis_queue_names)
    ) as executor:
        tasks_raw = executor.map(_fetch_func, redis_queue_names)
    return list(itertools.chain(*tasks_raw))


def _fetch_tasks_from_redis(
    r: redis.Redis, redis_queue_name: str
) -> List[QueuedTaskShort]:
    """Fetch tasks from redis."""
    tasks = []
    for obj_encoded in r.lrange(redis_queue_name, 0, -1):
        try:
            tasks.append(QueuedTaskShort.from_binary_json(obj_encoded))
        except ValueError:
            pass
    return reversed(tasks)


def delete_task_by_id(task_id: str) -> int:
    """Delete task from queue by task ID. Returns count of deleted entries."""
    return _delete_task_by_condition("id", task_id)


def delete_task_by_name(task_name: str) -> int:
    """Delete task from queue by task name. Returns count of deleted entries."""
    return _delete_task_by_condition("name", task_name)


def delete_task_by_app_name(app_name: str) -> int:
    """Delete task from queue by app name. Returns count of deleted entries."""
    return _delete_task_by_condition("app_name", app_name)


def _delete_task_by_condition(field: str, value) -> int:
    """Delete task from queue which match the condition.
    Returns count of deleted entries.
    """
    r = _redis_client()
    deleted_count = 0
    for queue_name in _redis_queue_names():
        for obj_encoded in r.lrange(queue_name, 0, -1):
            try:
                task = QueuedTaskShort.from_binary_json(obj_encoded)
            except ValueError:
                continue
            if getattr(task, field) == value:
                r.lrem(queue_name, 1, obj_encoded)
                deleted_count += 1
    return deleted_count


def clear_tasks(queue_name: str = None):
    """Clear tasks from all queues."""
    r = _redis_client()
    for redis_queue_name in _redis_queue_names(queue_name):
        r.delete(redis_queue_name)
    tasks_cache.clear()


def add_tasks(queue_name: str, raw_tasks: list):
    """Push fake tasks to task queue."""
    r = _redis_client()
    if not queue_name:
        queue_name = default_queue_name()
    tasks_by_priority = defaultdict(list)
    for task in raw_tasks:
        priority = task["properties"]["priority"]
        tasks_by_priority[priority].append(task)
    for priority, tasks in tasks_by_priority.items():
        raw_tasks_str = [json.dumps(obj) for obj in tasks]
        queue_name_raw = f"{queue_name}{PRIORITY_SEP}{priority}"
        r.lpush(queue_name_raw, *raw_tasks_str)
    del tasks_by_priority


tasks_cache = QueuedTasksCache(
    cache_key=QUEUED_TASKS_CACHE_KEY, timeout=TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT
)



"""Managers for Task Monitor."""

# pylint: disable = missing-class-docstring

import datetime as dt
import json
import traceback as tb
from typing import List, Optional
from uuid import UUID

from django.core.cache import cache
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Avg, Count, F, Max, Min, Q, Sum
from django.db.models.functions import TruncMinute
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from app_utils.database import TableSizeMixin
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT,
    TASKMONITOR_REPORTS_MAX_AGE,
    TASKMONITOR_TRUNCATE_NESTED_DATA,
)
from .core import celery_queues
from .helpers import extract_app_name, truncate_dict, truncate_list, truncate_result

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class QuerySetQueryStub:
    def __init__(self) -> None:
        self.select_related = None
        self.order_by = []


class ListAsQuerySet(list):
    """Masquerade a list as QuerySet."""

    def __init__(self, *args, model, distinct=False, **kwargs):
        self.model = model
        self.query = QuerySetQueryStub()
        self.distinct_enabled = distinct
        super().__init__(*args, **kwargs)
        self._id_mapper = {str(obj.id): n for n, obj in enumerate(self)}
        self._list_size = len(self)

    def all(self) -> models.QuerySet:
        """:private:"""
        return self

    def none(self) -> models.QuerySet:
        """:private:"""
        return ListAsQuerySet([], model=self.model)

    def get(self, *args, **kwargs):
        """:private:"""
        try:
            return self[self._id_mapper[str(kwargs["id"])]]
        except KeyError:
            raise self.model.DoesNotExist from None

    def distinct(self):
        """:private:"""
        return ListAsQuerySet(list(set(self)), model=self.model, distinct=True)

    def values(self, *args):
        """:private:"""
        result = [
            {k: v for k, v in obj.__dict__.items() if not args or k in args}
            for obj in self
        ]
        return result

    def values_list(self, *args, **kwargs):
        """:private:"""
        items = [tuple(obj.values()) for obj in self.values(*args)]
        if kwargs.get("flat"):
            items = [obj[0] for obj in items]
            if self.distinct_enabled:
                return list(dict.fromkeys(items))
            return items
        return items

    def first(self):
        """:private:"""
        return self[0] if self else None

    def filter(self, *args, **kwargs):
        """:private:"""
        if args:
            raise NotImplementedError("filter with positional args not supported.")
        if not kwargs:
            return self
        new_list = [
            obj
            for obj in self
            if all((str(getattr(obj, k)) == str(v) for k, v in kwargs.items()))
        ]
        return ListAsQuerySet(new_list, model=self.model)

    def order_by(self, *args, **kwargs):
        """:private:"""
        if kwargs:
            raise NotImplementedError("order with kw args not supported.")
        if args:
            for prop in reversed(args):
                if prop[0:1] == "-":
                    reverse = True
                    prop_2 = prop[1:]
                else:
                    reverse = False
                    prop_2 = prop
                # pylint: disable = cell-var-from-loop
                self.sort(key=lambda d: getattr(d, prop_2), reverse=reverse)
        return self

    def count(self):
        """:private:"""
        return self._list_size

    def _clone(self):
        return self


class QueuedTaskQuerySet(models.QuerySet):
    def count(self):
        """Return cached count."""
        return celery_queues.queue_length()


class QueuedTaskManagerBase(models.Manager):
    def get_queryset(self) -> models.QuerySet:
        """Return queryset generated from celery API."""
        if celery_queues.queue_length() > TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT:
            return self._none()
        return self.from_dto_list(celery_queues.fetch_tasks())

    @staticmethod
    def _none():
        from .models import QueuedTask

        return ListAsQuerySet([], model=QueuedTask)

    @staticmethod
    def from_dto_list(tasks: list) -> models.QuerySet:
        """Create from a list of QueuedTaskShort objects."""
        from .models import QueuedTask

        objs = [
            QueuedTask.from_dto(obj, position) for position, obj in enumerate(tasks)
        ]
        return ListAsQuerySet(objs, model=QueuedTask)


QueuedTaskManager = QueuedTaskManagerBase.from_queryset(QueuedTaskQuerySet)


class TaskLogQuerySet(models.QuerySet):
    def csv_line_generator(self, fields: List[str]):
        """Return the task logs for a CSV file line by line.
        And return the field names as first line.
        """
        field_names = [field.name for field in fields]
        yield field_names
        for obj in self.iterator():
            values = [
                value for key, value in obj.asdict().items() if key in set(field_names)
            ]
            yield values

    def aggregate_timestamp_trunc(self):
        """Aggregate timestamp trunc."""
        return (
            self.annotate(timestamp_trunc=TruncMinute("timestamp"))
            .values("timestamp_trunc")
            .annotate(task_runs=Count("id"))
        )

    def max_throughput(self) -> int:
        """Calculate the maximum throughput in task executions per minute."""
        qs = self.aggregate_timestamp_trunc().aggregate(Max("task_runs"))
        return qs["task_runs__max"]

    def avg_throughput(self) -> float:
        """Calculate the average throughput in task executions per minute."""
        qs = self.aggregate_timestamp_trunc().aggregate(Avg("task_runs"))
        return qs["task_runs__avg"]

    def oldest_date(self) -> dt.datetime:
        """Return oldest timestamp."""
        return self.aggregate(oldest=Min("timestamp"))["oldest"]

    def newest_date(self) -> dt.datetime:
        """Return newest timestamp."""
        return self.aggregate(youngest=Max("timestamp"))["youngest"]

    def filter_stale_logs(self, max_hours: int) -> models.QuerySet:
        """Filter stale logs."""
        deadline = timezone.now() - dt.timedelta(hours=max_hours)
        qs = self.filter(timestamp__lt=deadline)
        return qs


class TaskLogManagerBase(TableSizeMixin, models.Manager):
    # pylint: disable = too-many-locals, too-many-arguments
    def create_from_task(
        self,
        *,
        task_id: str,
        task_name: str,
        state: int,
        retries: int,
        priority: int,
        args: list,
        kwargs: dict,
        received: Optional[dt.datetime] = None,
        started: Optional[dt.datetime] = None,
        parent_id: Optional[str] = None,
        exception=None,
        result=None,
        current_queue_length: Optional[int] = None,
    ) -> models.Model:
        """Create new object from a celery task."""
        params = {
            "app_name": extract_app_name(task_name),
            "priority": priority,
            "parent_id": UUID(parent_id) if parent_id else None,
            "received": received,
            "retries": retries,
            "started": started,
            "state": state,
            "task_id": UUID(task_id),
            "task_name": task_name,
            "timestamp": timezone.now(),
            "current_queue_length": current_queue_length,
        }
        args = args or []
        params["args"] = (
            truncate_list(args) if TASKMONITOR_TRUNCATE_NESTED_DATA else args
        )
        kwargs = kwargs or {}
        params["kwargs"] = (
            truncate_dict(kwargs) if TASKMONITOR_TRUNCATE_NESTED_DATA else kwargs
        )
        try:
            json.dumps(result, cls=DjangoJSONEncoder)
        except TypeError:
            logger.warning(
                "%s [%s]: Result was not JSON serializable and therefore discarded.",
                task_name,
                task_id,
            )
            result = None
        params["result"] = (
            truncate_result(result) if TASKMONITOR_TRUNCATE_NESTED_DATA else result
        )
        if exception:
            params["exception"] = exception.__class__.__name__
            if traceback := getattr(exception, "__traceback__"):
                params["traceback"] = "".join(
                    tb.format_exception(None, value=exception, tb=traceback)
                )
        return self.create(**params)


TaskLogManager = TaskLogManagerBase.from_queryset(TaskLogQuerySet)


class TaskStatisticManager(models.Manager):
    _CACHE_KEY = "taskmonitor-task-statistics"
    _CACHE_TIMEOUT = TASKMONITOR_REPORTS_MAX_AGE * 60

    def get_queryset(self) -> models.QuerySet:
        """Return queryset with generated data from statistics query."""
        from .models import TaskStatistic

        objs = cache.get_or_set(
            key=self._CACHE_KEY, default=self._run_query, timeout=self._CACHE_TIMEOUT
        )

        return ListAsQuerySet(objs, model=TaskStatistic)

    @classmethod
    def refresh_cache(cls):
        """Update the query cache."""
        cache.set(
            key=cls._CACHE_KEY, value=cls._run_query(), timeout=cls._CACHE_TIMEOUT
        )

    @classmethod
    def clear_cache(cls):
        """Clear the query cache."""
        cache.delete(cls._CACHE_KEY)

    @classmethod
    def cached_at(cls) -> Optional[dt.datetime]:
        """Return datetime when cache was last created or None if there is no cache."""
        seconds = cache.ttl(cls._CACHE_KEY)
        if not seconds:
            return None
        return timezone.now() - (dt.timedelta(seconds=cls._CACHE_TIMEOUT - seconds))

    @staticmethod
    def _run_query() -> list:
        from .models import TaskLog, TaskStatistic

        excluded_fields = {"id"}
        field_names = [
            field.name
            for field in TaskStatistic._meta.get_fields()
            if field.name not in excluded_fields
        ]
        query = (
            TaskLog.objects.values(name=F("task_name"))
            .annotate(app=F("app_name"))
            .annotate(runs_total=Count("pk"))
            .annotate(runs_succeeded=Count("pk", filter=Q(state=TaskLog.State.SUCCESS)))
            .annotate(runs_failed=Count("pk", filter=Q(state=TaskLog.State.FAILURE)))
            .annotate(runs_retried=Count("pk", filter=Q(state=TaskLog.State.RETRY)))
            .annotate(runtime_min=Min("runtime"))
            .annotate(runtime_avg=Avg("runtime"))
            .annotate(runtime_max=Max("runtime"))
            .annotate(runtime_total=Sum("runtime"))
            .values(*field_names)
            .order_by("name")
        )
        items = [{**obj, **{"id": num}} for num, obj in enumerate(query, start=1)]
        objs = [TaskStatistic(**obj) for obj in items]
        return objs



"""Fake exceptions with traceback for tests."""


class FakeCode(object):
    """A fake code object for tests."""

    def __init__(self, co_filename, co_name):
        self.co_filename = co_filename
        self.co_name = co_name


class FakeFrame(object):
    """A fake frame for tests."""

    def __init__(self, f_code, f_globals):
        self.f_code = f_code
        self.f_globals = f_globals


class FakeTraceback(object):
    """A fake traceback for tests."""

    def __init__(self, frames, line_nums):
        if len(frames) != len(line_nums):
            raise ValueError("Ya messed up!")
        self._frames = frames
        self._line_nums = line_nums
        self.tb_frame = frames[0]
        self.tb_lineno = line_nums[0]
        self.tb_lasti = -1  # hack to make the tests work under Py 3.11

    @property
    def tb_next(self):
        if len(self._frames) > 1:
            return FakeTraceback(self._frames[1:], self._line_nums[1:])


class FakeException(Exception):
    """A fake exception for tests."""

    def __init__(self, *args, **kwargs):
        self._tb = None
        super().__init__(*args, **kwargs)

    @property
    def __traceback__(self):
        return self._tb

    @__traceback__.setter
    def __traceback__(self, value):
        self._tb = value

    def with_traceback(self, value):
        self._tb = value
        return self


def make_fake_exception():
    """Generate a fake exception and return it."""
    code1 = FakeCode("made_up_filename.py", "non_existent_function")
    code2 = FakeCode("another_non_existent_file.py", "another_non_existent_method")
    frame1 = FakeFrame(code1, {})
    frame2 = FakeFrame(code2, {})
    traceback = FakeTraceback([frame1, frame2], [1, 3])
    return FakeException("fake exception").with_traceback(traceback)



from collections import Counter
from enum import Enum

import humanize

from django.core.management.base import BaseCommand, CommandError

from taskmonitor import __title__, app_settings
from taskmonitor.core import cached_reports, celery_queues
from taskmonitor.models import TaskLog

CACHE_TIMEOUT_SECONDS = 3600


class Token(str, Enum):
    """Argument token."""

    COMMAND = "command"
    TARGET = "target"


class UserCommand(str, Enum):
    """Command for users."""

    PURGE = "purge"
    INSPECT = "inspect"


class Target(str, Enum):
    """Target of a command."""

    LOGS = "logs"
    QUEUE = "queue"
    SETTINGS = "settings"


def my_input(*args, **kwargs) -> str:
    """Helper to enable mocking of input for unit tests."""
    return input(*args, **kwargs)


class Command(BaseCommand):
    help = f"Command utility for {__title__}."

    def run_from_argv(self, *args, **kwargs) -> None:
        """Workaround to handle exceptions in missing sub commands more gracefully."""
        try:
            super().run_from_argv(*args, **kwargs)
        except CommandError as ex:
            token = __name__.split(".")
            try:
                name = token[-1:].pop()
            except IndexError:
                name = "?"
            print(f"manage.py {name}: {ex}")

    def add_arguments(self, parser) -> None:
        subparsers = parser.add_subparsers(
            dest=Token.COMMAND.value,
            required=True,
            title="commands",
            help="available commands",
        )
        # purge command
        parser_purge = subparsers.add_parser(
            UserCommand.PURGE.value, help="Purge a target"
        )
        parser_purge.add_argument(
            Token.TARGET.value,
            type=str,
            choices=[Target.QUEUE.value, Target.LOGS.value],
            help="target to purge",
        )
        purge_group = parser_purge.add_mutually_exclusive_group(required=True)
        purge_group.add_argument(
            "--all", action="store_true", help="You want to purge everything"
        )
        purge_group.add_argument(
            "--app-name", help="Limit purge to tasks from an specific app"
        )
        purge_group.add_argument(
            "--task-id", help="Limit purge to tasks with a specific ID"
        )
        purge_group.add_argument(
            "--task-name", help="Limit purge to tasks with a specific name"
        )

        # inspect command
        parser_inspect = subparsers.add_parser(
            UserCommand.INSPECT.value,
            help="Inspect a target, e.g. show information about it.",
        )
        parser_inspect.add_argument(
            Token.TARGET.value,
            type=str,
            choices=[Target.LOGS.value, Target.QUEUE.value, Target.SETTINGS.value],
            help="target to inspect",
        )
        parser_inspect.add_argument(
            "--force-calc",
            action="store_true",
            help="Fore re-calculation of values and update caches.",
        )

    def _user_confirmed(self, question_text):
        """Ask user about confirmation and exit
        when he does not reply in the affirmative.
        """
        user_input = my_input(f"{question_text} (y/N)?")
        if user_input.lower() != "y":
            self.stdout.write(self.style.WARNING("Aborted by user request."))
            exit(1)

    def _clear_line(self):
        self.stdout.write("" * 70, ending="\r")

    def purge_queue(self, options):
        num_entries = celery_queues.queue_length()
        if not num_entries:
            self.stdout.write(self.style.WARNING("Queue is empty. Aborted."))
            exit(1)
        self.stdout.write(f"Current queue size: {num_entries:,}")
        if options["task_name"]:
            task_name = options["task_name"]
            self._user_confirmed(
                f"Are you sure you purge all tasks with the TASK NAME {task_name} from the queue?"
            )
            self.stdout.write("Purged tasks from queue...", ending="\r")
            deleted_entries = celery_queues.delete_task_by_name(task_name)
            self._clear_line()
        elif options["task_id"]:
            task_id = options["task_id"]
            self._user_confirmed(
                f"Are you sure you purge all tasks with the TASK ID {task_id} from the queue?"
            )
            self.stdout.write("Purged tasks from queue...", ending="\r")
            deleted_entries = celery_queues.delete_task_by_id(task_id)
            self._clear_line()
        elif options["app_name"]:
            app_name = options["app_name"]
            self._user_confirmed(
                f"Are you sure you purge all tasks belonging to the app {app_name} from the queue?"
            )
            self.stdout.write("Purged tasks from queue...", ending="\r")
            deleted_entries = celery_queues.delete_task_by_app_name(app_name)
            self._clear_line()
        elif options["all"]:
            self._user_confirmed("Are you sure you purge ALL TASKS from the queue?")
            self.stdout.write("Purged tasks from queue...", ending="\r")
            celery_queues.clear_tasks()
            deleted_entries = num_entries
            self._clear_line()
        else:
            raise NotImplementedError("This option is not yet implemented")
        self.stdout.write(f"Purged {deleted_entries:,} tasks from queue...")
        self.stdout.write(self.style.SUCCESS("Done."))

    def purge_logs(self, options):
        all_logs = TaskLog.objects.all()
        self.stdout.write("Calculating...", ending="\r")
        num_logs = all_logs.count()
        if not num_logs:
            self.stdout.write(self.style.WARNING("No logs found. Aborted."))
            exit(1)
        self.stdout.write(f"Current task logs count: {num_logs:,}")
        if options["all"]:
            self._user_confirmed("Are you sure you purge ALL task logs?")
            self.stdout.write("Deleting task logs...")
            all_logs._raw_delete(all_logs.db)
        else:
            raise CommandError("Currently only the --all option is supported")
        cached_reports.refresh_cache()
        self.stdout.write(self.style.SUCCESS("Done."))

    def inspect_logs(self, options):
        log_count = TaskLog.objects.count()
        try:
            db_table_size = TaskLog.objects.db_table_size()
        except RuntimeError:
            table_size_str = "N/A"
            average_bytes_str = "N/A"
        else:
            table_size_str = humanize.naturalsize(db_table_size)
            average_bytes_str = (
                humanize.naturalsize(db_table_size / log_count) if log_count else "N/A"
            )
        output = {
            "Log count in DB": humanize.intword(log_count),
            "Table size in DB": table_size_str,
            "Average log size in DB": average_bytes_str,
        }
        max_length = max([len(o) for o in output.keys()])
        for label, value in output.items():
            self.stdout.write(f"{label:{max_length + 1}}: {value}")

    def inspect_queue(self, options):
        num_entries = celery_queues.queue_length()
        if not num_entries:
            self.stdout.write("Queue is empty.")
            return
        self.stdout.write(f"Current queue size: {num_entries:,}")
        if num_entries > app_settings.TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT:
            self._user_confirmed(
                "The queue is very large. Do you still want to gather statistics?"
            )
        self.stdout.write("Fetching data from queue...", ending="\r")
        self._clear_line()
        tasks = [
            (task.app_name, task.name)
            for task in celery_queues._fetch_task_from_all_queues()
        ]
        self.stdout.write("Queued tasks grouped by app counts in descending order:")
        grouped_names = (o[0] for o in tasks)
        self._render_field_counts(grouped_names)
        self.stdout.write("Queued tasks grouped by task counts in descending order:")
        grouped_names = (o[1] for o in tasks)
        self._render_field_counts(grouped_names)

    def _render_field_counts(self, grouped_names):
        field_counts = Counter(grouped_names)
        field_counts_sorted = dict(
            sorted(field_counts.items(), key=lambda item: item[1], reverse=True)
        )
        max_length = (
            max([len(o) for o in field_counts_sorted.keys()])
            if field_counts_sorted
            else 0
        )
        for app_name, count in field_counts_sorted.items():
            self.stdout.write(f"  {app_name:{max_length}}: {count:,}")

    def inspect_settings(self, options):
        settings = sorted(
            [o for o in dir(app_settings) if not o.startswith("__") and o == o.upper()]
        )
        max_length = max([len(o) for o in settings])
        for setting_name in settings:
            value = getattr(app_settings, setting_name)
            self.stdout.write(f"{setting_name:{max_length + 1}}: {value}")

    def handle(self, *args, **options):
        command = options[Token.COMMAND.value]
        target = options[Token.TARGET.value]

        method = f"{command}_{target}"
        try:
            getattr(self, method)(options)
        except AttributeError:
            raise NotImplementedError(method) from None



"""Admin site for Task Monitor."""

# pylint: disable = missing-class-docstring, missing-function-docstring

import datetime as dt
import json
from typing import Optional
from urllib.parse import urlencode

from django.contrib import admin
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils import html, safestring, timezone
from django.utils.translation import gettext_lazy as _

from app_utils.admin import FieldFilterCountsDb, FieldFilterCountsMemory

from .app_settings import (
    TASKMONITOR_DATA_MAX_AGE,
    TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT,
    TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT,
)
from .core import cached_reports, celery_queues
from .models import QueuedTask, TaskLog, TaskReport, TaskStatistic


class QueuedTaskAppsListFilter(FieldFilterCountsMemory):
    """Filter by app name and show name with counts."""

    title = _("app name")
    parameter_name = "app_name"
    field_name = "app_name"


class QueuedTaskTasksListFilter(FieldFilterCountsMemory):
    """Filter by task name and show name with counts."""

    title = _("task name")
    parameter_name = "task_name"
    field_name = "name"


@admin.register(QueuedTask)
class QueuedTaskAdmin(admin.ModelAdmin):
    list_display = (
        "position",
        "id",
        "name",
        "priority",
        "app_name",
    )
    list_display_links = None
    list_filter = (QueuedTaskAppsListFilter, "priority", QueuedTaskTasksListFilter)
    ordering = ["position"]

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        cache_created_at = celery_queues.tasks_cache.created_at() or timezone.now()
        objs_count = celery_queues.queue_length()
        context = {
            "title": "Currently queued tasks",
            "cache_created_at": cache_created_at,
            "task_count": objs_count,
            "cache_timeout": TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT,
            "is_below_limit": objs_count < TASKMONITOR_QUEUED_TASKS_ADMIN_LIMIT,
        }
        extra_context.update(context)
        return super().changelist_view(request, extra_context)


@admin.register(TaskReport)
class TaskReportAdmin(admin.ModelAdmin):
    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def changelist_view(self, request, extra_context=None):
        return redirect("taskmonitor:admin_taskmonitor_reports")


class TaskLogAppsListFilter(FieldFilterCountsDb):
    """Filter by app name and show name with counts."""

    title = _("app name")
    parameter_name = "app"
    field_name = "app_name"


class TaskLogStatesListFilter(FieldFilterCountsDb):
    """Filter by app name and show name with counts."""

    title = _("state")
    parameter_name = "state"
    field_name = "state"


class TaskLogExceptionsListFilter(FieldFilterCountsDb):
    """Filter by exception name and show name with counts."""

    title = _("exception")
    parameter_name = "exception"
    field_name = "exception"


@admin.register(TaskLog)
class TaskLogAdmin(admin.ModelAdmin):
    list_display = (
        "_timestamp",
        "task_name",
        "_params",
        "priority",
        "_state",
        "_runtime",
        "_exception",
    )
    list_filter = (
        TaskLogStatesListFilter,
        TaskLogExceptionsListFilter,
        "timestamp",
        TaskLogAppsListFilter,
        "priority",
    )
    search_fields = ("task_name", "app_name", "task_id")
    actions = ["delete_selected_2"]
    show_full_result_count = False
    fields = (
        "task_id",
        "task_name",
        "timestamp",
        "_args",
        "_kwargs",
        "_result",
        "retries",
        "priority",
        "state",
        "runtime",
        "app_name",
        "_exception",
        "parent_id",
        "received",
        "started",
        "_traceback",
    )

    def has_add_permission(self, *args, **kwargs) -> bool:
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def get_readonly_fields(self, request, obj=None):
        try:
            field = [f for f in TaskLog._meta.fields if f.name == "kwargs"]
            if len(field) > 0:
                field = field[0]
                field.help_text = "some special help text"
        except Exception:  # pylint: disable = broad-exception-caught
            pass
        return self.readonly_fields

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context.update(_task_log_stats())
        return super().changelist_view(request, extra_context)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        obj = get_object_or_404(TaskLog, pk=object_id)
        extra_context["tasklog_text"] = obj.astext()
        return super().change_view(
            request, object_id, form_url, extra_context=extra_context
        )

    @admin.display
    def _params(self, obj):
        if obj.args and not obj.kwargs:
            return html.format_html("<code>{}</code>", json.dumps(obj.args))
        if not obj.args and obj.kwargs:
            return html.format_html(
                "<code>{}</code>", json.dumps(obj.kwargs, sort_keys=True)
            )
        if obj.args and obj.kwargs:
            return html.format_html(
                "<code>{}<br>{}</code>",
                json.dumps(obj.args),
                json.dumps(obj.kwargs, sort_keys=True),
            )
        return None

    @admin.display(description="Exception")
    def _exception(self, obj) -> str:
        return html.format_html("<code>{}</code>", obj.exception)

    @admin.display(ordering="runtime")
    def _runtime(self, obj) -> Optional[str]:
        return f"{obj.runtime:.1f}" if obj.runtime else None

    @admin.display(ordering="state")
    def _state(self, obj: TaskLog) -> str:
        css_class_map = {
            TaskLog.State.RETRY: "state-retry",
            TaskLog.State.FAILURE: "text-danger",
        }
        css_class = css_class_map.get(obj.state, "")
        return html.format_html(
            '<span class="{}">{}</span>', css_class, obj.get_state_display()
        )

    @admin.display(ordering="timestamp")
    def _timestamp(self, obj: TaskLog) -> str:
        return obj.timestamp_formatted

    @admin.action(description="Delete selected entries (NO CONFIRMATION!")
    def delete_selected_2(self, request, queryset):
        entries_count = queryset.count()
        queryset._raw_delete(queryset.db)  # pylint: disable = protected-access
        self.message_user(request, f"Deleted {entries_count} entries.")

    @admin.display(description="Result")
    def _result(self, obj):
        if obj.state == TaskLog.State.SUCCESS:
            return _format_html_data(obj.result)
        return "-"

    @admin.display(description="Args")
    def _args(self, obj):
        return _format_html_data(obj.args)

    @admin.display(description="Kwargs")
    def _kwargs(self, obj):
        return _format_html_data(obj.kwargs)

    @admin.display(description="Traceback")
    def _traceback(self, obj):
        return _format_html_lines(obj.traceback)


@admin.register(TaskStatistic)
class TaskStatisticAdmin(admin.ModelAdmin):
    list_display = [
        "_name",
        "_runs_total",
        "_runs_succeeded",
        "_runs_failed",
        "_runs_retried",
        "_runtime_min",
        "_runtime_avg",
        "_runtime_max",
        "_runtime_total",
    ]
    list_display_links = None
    list_filter = ["app"]
    ordering = ["name"]

    @admin.display(ordering="name", description="name")
    def _name(self, obj: TaskStatistic):
        base_url = reverse("admin:taskmonitor_tasklog_changelist")
        query_encoded = urlencode({"task_name": obj.name})
        url = f"{base_url}?{query_encoded}"
        return html.format_html('<a href="{}">{}</a>', url, obj.name)

    @admin.display(ordering="runs_total", description="runs")
    def _runs_total(self, obj: TaskStatistic):
        return f"{obj.runs_total:,}"

    @admin.display(ordering="runs_succeeded", description="succeeded")
    def _runs_succeeded(self, obj: TaskStatistic):
        return f"{obj.runs_succeeded:,}"

    @admin.display(ordering="runs_failed", description="failed")
    def _runs_failed(self, obj: TaskStatistic):
        return f"{obj.runs_failed:,}"

    @admin.display(ordering="runs_retried", description="retried")
    def _runs_retried(self, obj: TaskStatistic):
        return f"{obj.runs_retried:,}"

    @admin.display(ordering="runtime_min")
    def _runtime_min(self, obj: TaskStatistic):
        return self._format_float(obj.runtime_min)

    @admin.display(ordering="runtime_avg")
    def _runtime_avg(self, obj: TaskStatistic):
        return self._format_float(obj.runtime_avg)

    @admin.display(ordering="runtime_max")
    def _runtime_max(self, obj: TaskStatistic):
        return self._format_float(obj.runtime_max)

    @admin.display(ordering="runtime_total")
    def _runtime_total(self, obj: TaskStatistic):
        return self._format_float(obj.runtime_total)

    @staticmethod
    def _format_float(value) -> str:
        return f"{value:,.2f}"

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def changelist_view(self, request, extra_context=None):
        context = {"title": "Task Statistics"}
        last_update_at = TaskStatistic.objects.cached_at()
        context.update(_task_log_stats(last_update_at))
        extra_context = extra_context or {}
        extra_context.update(context)
        return super().changelist_view(request, extra_context)


def _task_log_stats(last_update_at: Optional[dt.datetime] = None) -> dict:
    """Return dict with task log stats."""
    report = cached_reports.report("tasks_basics")
    report_data = report.data()
    total_runs = report_data["total_runs"]
    oldest_date = report_data["oldest_date"]
    newest_date = report_data["newest_date"]
    context = {
        "total_runs": total_runs,
        "oldest_date": oldest_date,
        "newest_date": newest_date,
        "data_max_age": TASKMONITOR_DATA_MAX_AGE,
        "last_update_at": last_update_at,
    }
    return context


def _format_html_lines(text) -> str:
    return safestring.mark_safe(
        "<br>".join(
            [html.format_html("<code>{}</code>", line) for line in text.splitlines()]
        )
    )


def _format_html_data(data) -> str:
    return html.format_html(
        "<code>{}</code>", json.dumps(data, sort_keys=True, indent=4)
    )



import json

from django.core.cache import cache
from django.test import TestCase

from taskmonitor.core.tasks_cache import QueuedTasksCache, QueuedTaskShort

from ..factories import QueuedTaskRawFactory


class TestQueuedTaskShort(TestCase):
    def test_should_create_from_dict(self):
        # given
        task = QueuedTaskRawFactory()
        # when
        obj = QueuedTaskShort.from_dict(task)
        # then
        self.assertEqual(obj.id, task["headers"]["id"])
        self.assertEqual(obj.name, task["headers"]["task"])
        self.assertEqual(obj.priority, task["properties"]["priority"])

    def test_should_create_from_redis_entry(self):
        # given
        task = QueuedTaskRawFactory()
        obj_json = json.dumps(task)
        obj_encoded = obj_json.encode("utf-8")
        # when
        obj = QueuedTaskShort.from_binary_json(obj_encoded)
        # then
        self.assertEqual(obj.id, task["headers"]["id"])
        self.assertEqual(obj.name, task["headers"]["task"])
        self.assertEqual(obj.priority, task["properties"]["priority"])


class TestCacheAPi(TestCase):
    CACHE_KEY = "test-cache-api-cache-key"

    def setUp(self) -> None:
        cache.delete(self.CACHE_KEY)

    def test_should_store_and_fetch_from_cache(self):
        # given
        tasks_cache = QueuedTasksCache(self.CACHE_KEY, 60)
        tasks = [QueuedTaskShort.from_dict(QueuedTaskRawFactory())]
        # when
        tasks_cache.set(tasks)
        result = tasks_cache.get()
        # then
        self.assertEqual(result.tasks, tasks)

    def test_should_raise_error_when_cache_returned_wrong_datatype(self):
        # given
        tasks_cache = QueuedTasksCache(self.CACHE_KEY)
        cache.set(key=self.CACHE_KEY, value="abc")
        # when/then
        with self.assertRaises(TypeError):
            tasks_cache.get()

    def test_should_clear_cache(self):
        # given
        tasks_cache = QueuedTasksCache(self.CACHE_KEY)
        tasks = [QueuedTaskShort.from_dict(QueuedTaskRawFactory())]
        tasks_cache.set(tasks)
        # when
        tasks_cache.clear()
        # then
        self.assertIsNone(tasks_cache.get())

    def test_should_return_created_at(self):
        # given
        tasks_cache = QueuedTasksCache(self.CACHE_KEY, 60)
        tasks = [QueuedTaskShort.from_dict(QueuedTaskRawFactory())]
        tasks_cache.set(tasks)
        data = tasks_cache.get()
        # when
        result = tasks_cache.created_at()
        # then
        self.assertEqual(result, data.created_at)

    def test_should_return_none_for_created_at_when_cache_invalid(self):
        # given
        tasks_cache = QueuedTasksCache(self.CACHE_KEY, 60)
        # when
        result = tasks_cache.created_at()
        # then
        self.assertIsNone(result)

    def test_should_disable_cache(self):
        # given
        tasks_cache = QueuedTasksCache(self.CACHE_KEY, 0)
        tasks = [QueuedTaskShort.from_dict(QueuedTaskRawFactory())]
        tasks_cache.set(tasks)
        # when
        result = tasks_cache.get()
        # then
        self.assertIsNone(result)



from unittest import mock

from django.test import TestCase

from taskmonitor.core import celery_queues
from taskmonitor.core.celery_queues import QueuedTaskShort, tasks_cache

from ..factories import QueuedTaskRawFactory

CELERY_QUEUE_NAME = "test_task_monitor_celery"
MODULE_PATH = "taskmonitor.core.celery_queues"


@mock.patch(MODULE_PATH + ".TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT", 10)
@mock.patch(MODULE_PATH + ".default_queue_name")
class TestCeleryQueues(TestCase):
    def setUp(self):
        celery_queues.clear_tasks(CELERY_QUEUE_NAME)

    def tearDown(self):
        celery_queues.clear_tasks(CELERY_QUEUE_NAME)

    def test_should_return_queue_length(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        raw_tasks = [QueuedTaskRawFactory(), QueuedTaskRawFactory()]
        celery_queues.add_tasks(CELERY_QUEUE_NAME, raw_tasks)
        # when/then
        self.assertEqual(celery_queues.queue_length(), 2)

    def test_should_clear_queue(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        raw_tasks = [QueuedTaskRawFactory(), QueuedTaskRawFactory()]
        celery_queues.add_tasks(CELERY_QUEUE_NAME, raw_tasks)
        # when
        celery_queues.clear_tasks()
        # then
        self.assertEqual(celery_queues.queue_length(), 0)

    def test_should_fetch_tasks_in_correct_order(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        raw_task_1 = QueuedTaskRawFactory(properties__priority=4)
        raw_task_2 = QueuedTaskRawFactory(properties__priority=4)
        raw_task_3 = QueuedTaskRawFactory(properties__priority=3)
        celery_queues.add_tasks(CELERY_QUEUE_NAME, [raw_task_1, raw_task_2, raw_task_3])
        # when
        with mock.patch(MODULE_PATH + ".tasks_cache") as m:
            m.get.return_value = None
            result = celery_queues.fetch_tasks()
        # then
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], QueuedTaskShort.from_dict(raw_task_3))
        self.assertEqual(result[1], QueuedTaskShort.from_dict(raw_task_1))
        self.assertEqual(result[2], QueuedTaskShort.from_dict(raw_task_2))

    def test_should_retrieve_tasks_from_cache(self, mock_queue_base_name):
        # given
        tasks = [QueuedTaskShort.from_dict(QueuedTaskRawFactory())]
        tasks_cache.set(tasks)
        # when
        result = celery_queues.fetch_tasks()
        # then
        self.assertEqual(result, tasks)

    def test_should_retrieve_tasks_from_redis_when_no_cache(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        tasks = [QueuedTaskRawFactory()]
        celery_queues.add_tasks(CELERY_QUEUE_NAME, tasks)
        # when
        result = celery_queues.fetch_tasks()
        # then
        self.assertEqual(result[0], QueuedTaskShort.from_dict(tasks[0]))

    def test_should_delete_task_by_id(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        task_1 = QueuedTaskRawFactory()
        task_2 = QueuedTaskRawFactory()
        celery_queues.add_tasks(CELERY_QUEUE_NAME, [task_1, task_2])
        task_1_dto = QueuedTaskShort.from_dict(task_1)
        task_2_dto = QueuedTaskShort.from_dict(task_2)
        # when
        result = celery_queues.delete_task_by_id(task_1_dto.id)
        # then
        self.assertEqual(result, 1)
        tasks = celery_queues.fetch_tasks()
        self.assertNotIn(task_1_dto, tasks)
        self.assertIn(task_2_dto, tasks)

    def test_should_delete_task_by_name(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        task_1 = QueuedTaskRawFactory(headers__task="alpha.tasks.blue")
        task_2 = QueuedTaskRawFactory(headers__task="alpha.tasks.yellow")
        celery_queues.add_tasks(CELERY_QUEUE_NAME, [task_1, task_2])
        task_1_dto = QueuedTaskShort.from_dict(task_1)
        task_2_dto = QueuedTaskShort.from_dict(task_2)
        # when
        result = celery_queues.delete_task_by_name(task_1_dto.name)
        # then
        self.assertEqual(result, 1)
        tasks = celery_queues.fetch_tasks()
        self.assertNotIn(task_1_dto, tasks)
        self.assertIn(task_2_dto, tasks)

    def test_should_delete_task_by_app_name(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        task_1 = QueuedTaskRawFactory(headers__task="alpha.tasks.blue")
        task_2 = QueuedTaskRawFactory(headers__task="bravo.tasks.yellow")
        celery_queues.add_tasks(CELERY_QUEUE_NAME, [task_1, task_2])
        task_1_dto = QueuedTaskShort.from_dict(task_1)
        task_2_dto = QueuedTaskShort.from_dict(task_2)
        # when
        result = celery_queues.delete_task_by_app_name(task_1_dto.app_name)
        # then
        self.assertEqual(result, 1)
        tasks = celery_queues.fetch_tasks()
        self.assertNotIn(task_1_dto, tasks)
        self.assertIn(task_2_dto, tasks)



"""API for caching queued tasks."""

import datetime as dt
import json
from dataclasses import dataclass
from typing import List, NamedTuple, Optional

from django.core.cache import cache
from django.utils import timezone

from taskmonitor.helpers import extract_app_name


class QueuedTaskShort(NamedTuple):
    """DTO for queued tasks, optimized for size."""

    app_name: str
    id: int
    name: str
    priority: int

    @classmethod
    def from_dict(cls, obj: dict) -> "QueuedTaskShort":
        """Create QueuedTaskShort from raw task dict."""
        if "headers" not in obj:
            raise ValueError("headers missing in obj")
        headers = obj["headers"]
        properties = obj["properties"] if "properties" in obj else {}
        task_name = headers["task"]
        return cls(
            app_name=extract_app_name(task_name),
            id=headers["id"],
            name=task_name,
            priority=properties.get("priority"),
        )

    @classmethod
    def from_binary_json(cls, obj_encoded: bytes) -> "QueuedTaskShort":
        """Create new object from a binary task object retrieved from Redis."""
        obj = json.loads(obj_encoded.decode("utf8"))
        return cls.from_dict(obj)


class QueuedTaskCacheEntry(NamedTuple):
    """Wrapper for storing queued tasks in cache with additional data."""

    tasks: List[QueuedTaskShort]
    created_at: dt.datetime


@dataclass(frozen=True)
class QueuedTasksCache:
    """Cache for queued tasks.

    Can be disabled when timeout is set to 0.

    Args:
        - cache_key: custom string used to store the cache
        - timeout: timeout for cache in seconds
    """

    cache_key: str
    timeout: int = 60

    def get(self) -> Optional[QueuedTaskCacheEntry]:
        """Retrieve content of cache or None if cache is expired or invalid."""
        if not self.timeout:
            return None
        data = cache.get(self.cache_key)
        if data is not None and not isinstance(data, QueuedTaskCacheEntry):
            raise TypeError("Expected data to be of type 'QueuedTaskCacheEntry'")
        return data

    def set(self, tasks: List[QueuedTaskShort]) -> Optional[QueuedTaskCacheEntry]:
        """Store given tasks in cache."""
        if not self.timeout:
            return None
        data = QueuedTaskCacheEntry(tasks=tasks, created_at=timezone.now())
        cache.set(self.cache_key, data, timeout=self.timeout)
        return data

    def clear(self):
        """Clear the cache."""
        cache.delete(self.cache_key)

    def created_at(self) -> Optional[dt.datetime]:
        """Return date when cache was created or None when cache is invalid."""
        data = self.get()
        if data is None:
            return None
        return data.created_at



"""Helpers for Task Monitor."""

# pylint: disable = protected-access

import datetime as dt
import functools


class Echo:
    """An object that implements just the write method of the file-like
    interface.
    """

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


def extract_app_name(task_name: str) -> str:
    """Extract the app name from a typical task name."""
    parts = task_name.split(".")
    try:
        idx = parts.index("tasks")
    except ValueError:
        if len(parts) == 2:
            return parts[0]
        return ""
    return parts[idx - 1] if idx > 0 else ""


def dict_sort_keys(dct: dict) -> dict:
    """Return a copy of this dictionary with sorted keys."""
    return dict(sorted(dct.items(), key=lambda x: x[0].lower()))


def truncate_list(lst: list) -> list:
    """Truncate nested elements and return as new list.

    Dicts will be replaced by `{}`

    Lists, tuple and sets wil be replaced by `[]`
    """
    return [_replace_nested_element(item) for item in lst]


def truncate_dict(dct: dict) -> dict:
    """Truncate nested values and return as new dict.

    Example:
    `{"a": {"aa": 1, ...}, "b": [1, 2, ...]}`-> `{"a": {}, "b": []}`
    """
    return {key: _replace_nested_element(value) for key, value in dct.items()}


def _replace_nested_element(value):
    if isinstance(value, dict):
        return {}
    if isinstance(value, (list, tuple, set)):
        return []
    return value


def truncate_result(value):
    """Truncate nested items in results and return as new value."""
    if isinstance(value, dict):
        return truncate_dict(value)
    if isinstance(value, (list, tuple, set)):
        return compress_list(truncate_list(value))
    return value


def compress_list(lst: list) -> list:
    """Compress list to empty list of it contains of empty containers only.

    Example: `[ {}, {} ]` -> `[]`
    """
    for item in lst:
        if item is False or item:
            return lst
    return []


def memcached(timeout: int = 30):
    """Cache result of decorated function in memory until timeout.

    Args:
    - timeout: Seconds until cache becomes stale
    """

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                my_cache = wrapper._cache
            except AttributeError:
                my_cache = None
            if (
                not my_cache
                or (dt.datetime.utcnow() - my_cache["timestamp"]).total_seconds()
                > timeout
            ):
                wrapper._cache = my_cache = {
                    "timestamp": dt.datetime.utcnow(),
                    "result": func(*args, **kwargs),
                }
            return my_cache["result"]

        return wrapper

    return actual_decorator



"""Tasks for Task Monitor."""

from celery import shared_task

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import TASKMONITOR_DATA_MAX_AGE, TASKMONITOR_DELETE_STALE_BATCH_SIZE
from .core import cached_reports
from .models import TaskLog, TaskStatistic

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

DEFAULT_TASK_PRIORITY = 4


@shared_task(base=QueueOnce)
def run_housekeeping():
    """Run all housekeeping tasks."""
    delete_stale_tasklogs.apply_async(priority=DEFAULT_TASK_PRIORITY)
    refresh_cached_data.apply_async(priority=DEFAULT_TASK_PRIORITY)


# @shared_task
# def delete_stale_tasklogs():
#     """Delete all stale tasklogs from the database."""
#     log_pks = TaskLog.objects.filter(
#         timestamp__lte=timezone.now() - dt.timedelta(hours=TASKMONITOR_DATA_MAX_AGE)
#     ).values_list("pk", flat=True)
#     for log_pks_chunk in chunks(log_pks, TASKMONITOR_DELETE_STALE_BATCH_SIZE):
#         delete_tasklogs_batch.apply_async(priority=7, kwargs={"log_pks": log_pks_chunk})


# @shared_task
# def delete_tasklogs_batch(log_pks: list):
#     """Delete a selection of tasklogs."""
#     logs_to_delete = TaskLog.objects.filter(pk__in=log_pks)
#     logger.info(f"Deleting {logs_to_delete.count():,} stale tasklogs.")
#     logs_to_delete._raw_delete(logs_to_delete.db)


@shared_task(base=QueueOnce)
def delete_stale_tasklogs():
    """Delete stale logs in batches.

    Will spawn itself again and again until all stale logs are deleted.
    """

    stale_logs = TaskLog.objects.filter_stale_logs(max_hours=TASKMONITOR_DATA_MAX_AGE)
    if not stale_logs.exists():
        logger.info("There currently are no stale application logs.")
        return

    stale_logs_pks = list(stale_logs.values_list("pk", flat=True))
    for pks_to_delete in chunks(stale_logs_pks, TASKMONITOR_DELETE_STALE_BATCH_SIZE):
        qs_to_delete = TaskLog.objects.filter(pk__in=pks_to_delete)
        qs_to_delete._raw_delete(qs_to_delete.db)  # pylint: disable = protected-access
        logger.info("Deleted %d stale task logs.", len(pks_to_delete))


@shared_task
def refresh_cached_data():
    """Refresh cached data incl. reports."""
    refresh_statistics_cache.apply_async(priority=DEFAULT_TASK_PRIORITY)

    for report in cached_reports.reports():
        refresh_single_report_cache.apply_async(
            priority=DEFAULT_TASK_PRIORITY, args=[report.name]
        )


@shared_task(base=QueueOnce)
def refresh_single_report_cache(report_name: str):
    """Refresh cache for given report."""
    cached_reports.report(report_name).refresh_cache()
    logger.info(f"Refreshed reports cache of {report_name}.")


@shared_task(base=QueueOnce)
def refresh_statistics_cache():
    """Refresh cache for given report."""
    TaskStatistic.objects.refresh_cache()
    logger.info("Refreshed statistics cache.")
